# -*- coding: utf-8 -*-

from typing import Union, Callable

import brainpy.math as bm
from brainpy.dyn.base import NeuGroup
from brainpy.initialize import ZeroInit, OneInit, Initializer, init_param
from brainpy.integrators.joint_eq import JointEq
from brainpy.integrators.ode import odeint
from brainpy.tools.checking import check_initializer
from brainpy.types import Shape, Tensor

__all__ = [
  'LIF',
  'ExpIF',
  'AdExIF',
  'QuaIF',
  'AdQuaIF',
  'GIF',
  'Izhikevich',
  'HindmarshRose',
  'FHN',
]


class LIF(NeuGroup):
  r"""Leaky integrate-and-fire neuron model.

  **Model Descriptions**

  The formal equations of a LIF model [1]_ is given by:

  .. math::

      \tau \frac{dV}{dt} = - (V(t) - V_{rest}) + I(t) \\
      \text{after} \quad V(t) \gt V_{th}, V(t) = V_{reset} \quad
      \text{last} \quad \tau_{ref} \quad  \text{ms}

  where :math:`V` is the membrane potential, :math:`V_{rest}` is the resting
  membrane potential, :math:`V_{reset}` is the reset membrane potential,
  :math:`V_{th}` is the spike threshold, :math:`\tau` is the time constant,
  :math:`\tau_{ref}` is the refractory time period,
  and :math:`I` is the time-variant synaptic inputs.

  **Model Examples**

  - `(Brette, Romain. 2004) LIF phase locking <https://brainpy-examples.readthedocs.io/en/latest/neurons/Romain_2004_LIF_phase_locking.html>`_

  **Model Parameters**

  ============= ============== ======== =========================================
  **Parameter** **Init Value** **Unit** **Explanation**
  ------------- -------------- -------- -----------------------------------------
  V_rest         0              mV       Resting membrane potential.
  V_reset        -5             mV       Reset potential after spike.
  V_th           20             mV       Threshold potential of spike.
  tau            10             ms       Membrane time constant. Compute by R * C.
  tau_ref       5              ms       Refractory period length.(ms)
  ============= ============== ======== =========================================

  **Neuron Variables**

  ================== ================= =========================================================
  **Variables name** **Initial Value** **Explanation**
  ------------------ ----------------- ---------------------------------------------------------
  V                    0                Membrane potential.
  input                0                External and synaptic input current.
  spike                False             Flag to mark whether the neuron is spiking.
  refractory           False             Flag to mark whether the neuron is in refractory period.
  t_last_spike         -1e7              Last spike time stamp.
  ================== ================= =========================================================

  **References**

  .. [1] Abbott, Larry F. "Lapicque’s introduction of the integrate-and-fire model
         neuron (1907)." Brain research bulletin 50, no. 5-6 (1999): 303-304.
  """

  def __init__(
      self,
      size: Shape,
      V_rest: Union[float, Tensor, Initializer, Callable] = 0.,
      V_reset: Union[float, Tensor, Initializer, Callable] = -5.,
      V_th: Union[float, Tensor, Initializer, Callable] = 20.,
      tau: Union[float, Tensor, Initializer, Callable] = 10.,
      tau_ref: Union[float, Tensor, Initializer, Callable] = 1.,
      V_initializer: Union[Initializer, Callable, Tensor] = ZeroInit(),
      method: str = 'exp_auto',
      name: str = None
  ):
    # initialization
    super(LIF, self).__init__(size=size, name=name)

    # parameters
    self.V_rest = init_param(V_rest, self.num, allow_none=False)
    self.V_reset = init_param(V_reset, self.num, allow_none=False)
    self.V_th = init_param(V_th, self.num, allow_none=False)
    self.tau = init_param(tau, self.num, allow_none=False)
    self.tau_ref = init_param(tau_ref, self.num, allow_none=False)

    # variables
    check_initializer(V_initializer, 'V_initializer')
    self.V = bm.Variable(init_param(V_initializer, (self.num,)))
    self.input = bm.Variable(bm.zeros(self.num))
    self.spike = bm.Variable(bm.zeros(self.num, dtype=bool))
    self.t_last_spike = bm.Variable(bm.ones(self.num) * -1e7)
    self.refractory = bm.Variable(bm.zeros(self.num, dtype=bool))

    # integral
    self.integral = odeint(method=method, f=self.derivative)

  def derivative(self, V, t, I_ext):
    dvdt = (-V + self.V_rest + I_ext) / self.tau
    return dvdt

  def update(self, _t, _dt):
    refractory = (_t - self.t_last_spike) <= self.tau_ref
    V = self.integral(self.V, _t, self.input, dt=_dt)
    V = bm.where(refractory, self.V, V)
    spike = V >= self.V_th
    self.t_last_spike.value = bm.where(spike, _t, self.t_last_spike)
    self.V.value = bm.where(spike, self.V_reset, V)
    self.refractory.value = bm.logical_or(refractory, spike)
    self.spike.value = spike
    self.input[:] = 0.


class ExpIF(NeuGroup):
  r"""Exponential integrate-and-fire neuron model.

  **Model Descriptions**

  In the exponential integrate-and-fire model [1]_, the differential
  equation for the membrane potential is given by

  .. math::

      \tau\frac{d V}{d t}= - (V-V_{rest}) + \Delta_T e^{\frac{V-V_T}{\Delta_T}} + RI(t), \\
      \text{after} \, V(t) \gt V_{th}, V(t) = V_{reset} \, \text{last} \, \tau_{ref} \, \text{ms}

  This equation has an exponential nonlinearity with "sharpness" parameter :math:`\Delta_{T}`
  and "threshold" :math:`\vartheta_{rh}`.

  The moment when the membrane potential reaches the numerical threshold :math:`V_{th}`
  defines the firing time :math:`t^{(f)}`. After firing, the membrane potential is reset to
  :math:`V_{rest}` and integration restarts at time :math:`t^{(f)}+\tau_{\rm ref}`,
  where :math:`\tau_{\rm ref}` is an absolute refractory time.
  If the numerical threshold is chosen sufficiently high, :math:`V_{th}\gg v+\Delta_T`,
  its exact value does not play any role. The reason is that the upswing of the action
  potential for :math:`v\gg v +\Delta_{T}` is so rapid, that it goes to infinity in
  an incredibly short time. The threshold :math:`V_{th}` is introduced mainly for numerical
  convenience. For a formal mathematical analysis of the model, the threshold can be pushed
  to infinity.

  The model was first introduced by Nicolas Fourcaud-Trocmé, David Hansel, Carl van Vreeswijk
  and Nicolas Brunel [1]_. The exponential nonlinearity was later confirmed by Badel et al. [3]_.
  It is one of the prominent examples of a precise theoretical prediction in computational
  neuroscience that was later confirmed by experimental neuroscience.

  Two important remarks:

  - (i) The right-hand side of the above equation contains a nonlinearity
    that can be directly extracted from experimental data [3]_. In this sense the exponential
    nonlinearity is not an arbitrary choice but directly supported by experimental evidence.
  - (ii) Even though it is a nonlinear model, it is simple enough to calculate the firing
    rate for constant input, and the linear response to fluctuations, even in the presence
    of input noise [4]_.

  **Model Examples**

  .. plot::
    :include-source: True

    >>> import brainpy as bp
    >>> group = bp.dyn.ExpIF(1)
    >>> runner = bp.dyn.DSRunner(group, monitors=['V'], inputs=('input', 10.))
    >>> runner.run(300., )
    >>> bp.visualize.line_plot(runner.mon.ts, runner.mon.V, ylabel='V', show=True)


  **Model Parameters**

  ============= ============== ======== ===================================================
  **Parameter** **Init Value** **Unit** **Explanation**
  ------------- -------------- -------- ---------------------------------------------------
  V_rest        -65            mV       Resting potential.
  V_reset       -68            mV       Reset potential after spike.
  V_th          -30            mV       Threshold potential of spike.
  V_T           -59.9          mV       Threshold potential of generating action potential.
  delta_T       3.48           \        Spike slope factor.
  R             1              \        Membrane resistance.
  tau           10             \        Membrane time constant. Compute by R * C.
  tau_ref       1.7            \        Refractory period length.
  ============= ============== ======== ===================================================

  **Model Variables**

  ================== ================= =========================================================
  **Variables name** **Initial Value** **Explanation**
  ------------------ ----------------- ---------------------------------------------------------
  V                  0                 Membrane potential.
  input              0                 External and synaptic input current.
  spike              False             Flag to mark whether the neuron is spiking.
  refractory         False             Flag to mark whether the neuron is in refractory period.
  t_last_spike       -1e7              Last spike time stamp.
  ================== ================= =========================================================

  **References**

  .. [1] Fourcaud-Trocmé, Nicolas, et al. "How spike generation
         mechanisms determine the neuronal response to fluctuating
         inputs." Journal of Neuroscience 23.37 (2003): 11628-11640.
  .. [2] Gerstner, W., Kistler, W. M., Naud, R., & Paninski, L. (2014).
         Neuronal dynamics: From single neurons to networks and models
         of cognition. Cambridge University Press.
  .. [3] Badel, Laurent, Sandrine Lefort, Romain Brette, Carl CH Petersen,
         Wulfram Gerstner, and Magnus JE Richardson. "Dynamic IV curves
         are reliable predictors of naturalistic pyramidal-neuron voltage
         traces." Journal of Neurophysiology 99, no. 2 (2008): 656-666.
  .. [4] Richardson, Magnus JE. "Firing-rate response of linear and nonlinear
         integrate-and-fire neurons to modulated current-based and
         conductance-based synaptic drive." Physical Review E 76, no. 2 (2007): 021919.
  .. [5] https://en.wikipedia.org/wiki/Exponential_integrate-and-fire
  """

  def __init__(
      self,
      size: Shape,
      V_rest: Union[float, Tensor, Initializer, Callable] = -65.,
      V_reset: Union[float, Tensor, Initializer, Callable] = -68.,
      V_th: Union[float, Tensor, Initializer, Callable] = -30.,
      V_T: Union[float, Tensor, Initializer, Callable] = -59.9,
      delta_T: Union[float, Tensor, Initializer, Callable] = 3.48,
      R: Union[float, Tensor, Initializer, Callable] = 1.,
      tau: Union[float, Tensor, Initializer, Callable] = 10.,
      tau_ref: Union[float, Tensor, Initializer, Callable] = 1.7,
      V_initializer: Union[Initializer, Callable, Tensor] = ZeroInit(),
      method: str = 'exp_auto',
      name: str = None
  ):
    # initialize
    super(ExpIF, self).__init__(size=size, name=name)

    # parameters
    self.V_rest = init_param(V_rest, self.num, allow_none=False)
    self.V_reset = init_param(V_reset, self.num, allow_none=False)
    self.V_th = init_param(V_th, self.num, allow_none=False)
    self.V_T = init_param(V_T, self.num, allow_none=False)
    self.delta_T = init_param(delta_T, self.num, allow_none=False)
    self.tau_ref = init_param(tau_ref, self.num, allow_none=False)
    self.tau = init_param(tau, self.num, allow_none=False)
    self.R = init_param(R, self.num, allow_none=False)

    # variables
    check_initializer(V_initializer, 'V_initializer')
    self.V = bm.Variable(init_param(V_initializer, (self.num,)))
    self.input = bm.Variable(bm.zeros(self.num))
    self.spike = bm.Variable(bm.zeros(self.num, dtype=bool))
    self.refractory = bm.Variable(bm.zeros(self.num, dtype=bool))
    self.t_last_spike = bm.Variable(bm.ones(self.num) * -1e7)

    # integral
    self.integral = odeint(method=method, f=self.derivative)

  def derivative(self, V, t, I_ext):
    exp_v = self.delta_T * bm.exp((V - self.V_T) / self.delta_T)
    dvdt = (- (V - self.V_rest) + exp_v + self.R * I_ext) / self.tau
    return dvdt

  def update(self, _t, _dt):
    refractory = (_t - self.t_last_spike) <= self.tau_ref
    V = self.integral(self.V, _t, self.input, dt=_dt)
    V = bm.where(refractory, self.V, V)
    spike = self.V_th <= V
    self.t_last_spike.value = bm.where(spike, _t, self.t_last_spike)
    self.V.value = bm.where(spike, self.V_reset, V)
    self.refractory.value = bm.logical_or(refractory, spike)
    self.spike.value = spike
    self.input[:] = 0.


class AdExIF(NeuGroup):
  r"""Adaptive exponential integrate-and-fire neuron model.

  **Model Descriptions**

  The **adaptive exponential integrate-and-fire model**, also called AdEx, is a
  spiking neuron model with two variables [1]_ [2]_.

  .. math::

      \begin{aligned}
      \tau_m\frac{d V}{d t} &= - (V-V_{rest}) + \Delta_T e^{\frac{V-V_T}{\Delta_T}} - Rw + RI(t), \\
      \tau_w \frac{d w}{d t} &=a(V-V_{rest}) - w
      \end{aligned}

  once the membrane potential reaches the spike threshold,

  .. math::

      V \rightarrow V_{reset}, \\
      w \rightarrow w+b.

  The first equation describes the dynamics of the membrane potential and includes
  an activation term with an exponential voltage dependence. Voltage is coupled to
  a second equation which describes adaptation. Both variables are reset if an action
  potential has been triggered. The combination of adaptation and exponential voltage
  dependence gives rise to the name Adaptive Exponential Integrate-and-Fire model.

  The adaptive exponential integrate-and-fire model is capable of describing known
  neuronal firing patterns, e.g., adapting, bursting, delayed spike initiation,
  initial bursting, fast spiking, and regular spiking.

  **Model Examples**

  - `Examples for different firing patterns <https://brainpy-examples.readthedocs.io/en/latest/neurons/Gerstner_2005_AdExIF_model.html>`_

  **Model Parameters**

  ============= ============== ======== ========================================================================================================================
  **Parameter** **Init Value** **Unit** **Explanation**
  ------------- -------------- -------- ------------------------------------------------------------------------------------------------------------------------
  V_rest        -65            mV       Resting potential.
  V_reset       -68            mV       Reset potential after spike.
  V_th          -30            mV       Threshold potential of spike and reset.
  V_T           -59.9          mV       Threshold potential of generating action potential.
  delta_T       3.48           \        Spike slope factor.
  a             1              \        The sensitivity of the recovery variable :math:`u` to the sub-threshold fluctuations of the membrane potential :math:`v`
  b             1              \        The increment of :math:`w` produced by a spike.
  R             1              \        Membrane resistance.
  tau           10             ms       Membrane time constant. Compute by R * C.
  tau_w         30             ms       Time constant of the adaptation current.
  ============= ============== ======== ========================================================================================================================

  **Model Variables**

  ================== ================= =========================================================
  **Variables name** **Initial Value** **Explanation**
  ------------------ ----------------- ---------------------------------------------------------
  V                   0                 Membrane potential.
  w                   0                 Adaptation current.
  input               0                 External and synaptic input current.
  spike               False             Flag to mark whether the neuron is spiking.
  t_last_spike        -1e7              Last spike time stamp.
  ================== ================= =========================================================

  **References**

  .. [1] Fourcaud-Trocmé, Nicolas, et al. "How spike generation
         mechanisms determine the neuronal response to fluctuating
         inputs." Journal of Neuroscience 23.37 (2003): 11628-11640.
  .. [2] http://www.scholarpedia.org/article/Adaptive_exponential_integrate-and-fire_model
  """

  def __init__(
      self,
      size: Shape,
      V_rest: Union[float, Tensor, Initializer, Callable] = -65.,
      V_reset: Union[float, Tensor, Initializer, Callable] = -68.,
      V_th: Union[float, Tensor, Initializer, Callable] = -30.,
      V_T: Union[float, Tensor, Initializer, Callable] = -59.9,
      delta_T: Union[float, Tensor, Initializer, Callable] = 3.48,
      a: Union[float, Tensor, Initializer, Callable] = 1.,
      b: Union[float, Tensor, Initializer, Callable] = 1.,
      tau: Union[float, Tensor, Initializer, Callable] = 10.,
      tau_w: Union[float, Tensor, Initializer, Callable] = 30.,
      R: Union[float, Tensor, Initializer, Callable] = 1.,
      V_initializer: Union[Initializer, Callable, Tensor] = ZeroInit(),
      w_initializer: Union[Initializer, Callable, Tensor] = ZeroInit(),
      method: str = 'exp_auto',
      name: str = None
  ):
    super(AdExIF, self).__init__(size=size, name=name)

    # parameters
    self.V_rest = init_param(V_rest, self.num, allow_none=False)
    self.V_reset = init_param(V_reset, self.num, allow_none=False)
    self.V_th = init_param(V_th, self.num, allow_none=False)
    self.V_T = init_param(V_T, self.num, allow_none=False)
    self.delta_T = init_param(delta_T, self.num, allow_none=False)
    self.a = init_param(a, self.num, allow_none=False)
    self.b = init_param(b, self.num, allow_none=False)
    self.tau = init_param(tau, self.num, allow_none=False)
    self.tau_w = init_param(tau_w, self.num, allow_none=False)
    self.R = init_param(R, self.num, allow_none=False)

    # variables
    check_initializer(V_initializer, 'V_initializer')
    check_initializer(w_initializer, 'w_initializer')
    self.V = bm.Variable(init_param(V_initializer, (self.num,)))
    self.w = bm.Variable(init_param(w_initializer, (self.num,)))
    self.refractory = bm.Variable(bm.zeros(self.num, dtype=bool))
    self.input = bm.Variable(bm.zeros(self.num))
    self.spike = bm.Variable(bm.zeros(self.num, dtype=bool))
    self.t_last_spike = bm.Variable(bm.ones(self.num) * -1e7)

    # functions
    self.integral = odeint(method=method, f=self.derivative)

  def dV(self, V, t, w, I_ext):
    dVdt = (- V + self.V_rest + self.delta_T * bm.exp((V - self.V_T) / self.delta_T) -
            self.R * w + self.R * I_ext) / self.tau
    return dVdt

  def dw(self, w, t, V):
    dwdt = (self.a * (V - self.V_rest) - w) / self.tau_w
    return dwdt

  @property
  def derivative(self):
    return JointEq([self.dV, self.dw])

  def update(self, _t, _dt):
    V, w = self.integral(self.V, self.w, _t, self.input, dt=_dt)
    spike = V >= self.V_th
    self.t_last_spike[:] = bm.where(spike, _t, self.t_last_spike)
    self.V.value = bm.where(spike, self.V_reset, V)
    self.w.value = bm.where(spike, w + self.b, w)
    self.spike.value = spike
    self.input[:] = 0.


class QuaIF(NeuGroup):
  r"""Quadratic Integrate-and-Fire neuron model.

  **Model Descriptions**

  In contrast to physiologically accurate but computationally expensive
  neuron models like the Hodgkin–Huxley model, the QIF model [1]_ seeks only
  to produce **action potential-like patterns** and ignores subtleties
  like gating variables, which play an important role in generating action
  potentials in a real neuron. However, the QIF model is incredibly easy
  to implement and compute, and relatively straightforward to study and
  understand, thus has found ubiquitous use in computational neuroscience.

  .. math::

      \tau \frac{d V}{d t}=c(V-V_{rest})(V-V_c) + RI(t)

  where the parameters are taken to be :math:`c` =0.07, and :math:`V_c = -50 mV` (Latham et al., 2000).

  **Model Examples**

  .. plot::
    :include-source: True

    >>> import brainpy as bp
    >>>
    >>> group = bp.dyn.QuaIF(1,)
    >>>
    >>> runner = bp.dyn.DSRunner(group, monitors=['V'], inputs=('input', 20.))
    >>> runner.run(duration=200.)
    >>> bp.visualize.line_plot(runner.mon.ts, runner.mon.V, show=True)


  **Model Parameters**

  ============= ============== ======== ========================================================================================================================
  **Parameter** **Init Value** **Unit** **Explanation**
  ------------- -------------- -------- ------------------------------------------------------------------------------------------------------------------------
  V_rest        -65            mV       Resting potential.
  V_reset       -68            mV       Reset potential after spike.
  V_th          -30            mV       Threshold potential of spike and reset.
  V_c           -50            mV       Critical voltage for spike initiation. Must be larger than V_rest.
  c             .07            \        Coefficient describes membrane potential update. Larger than 0.
  R             1              \        Membrane resistance.
  tau           10             ms       Membrane time constant. Compute by R * C.
  tau_ref       0              ms       Refractory period length.
  ============= ============== ======== ========================================================================================================================

  **Model Variables**

  ================== ================= =========================================================
  **Variables name** **Initial Value** **Explanation**
  ------------------ ----------------- ---------------------------------------------------------
  V                   0                 Membrane potential.
  input               0                 External and synaptic input current.
  spike               False             Flag to mark whether the neuron is spiking.
  refractory          False             Flag to mark whether the neuron is in refractory period.
  t_last_spike       -1e7               Last spike time stamp.
  ================== ================= =========================================================

  **References**

  .. [1]  P. E. Latham, B.J. Richmond, P. Nelson and S. Nirenberg
          (2000) Intrinsic dynamics in neuronal networks. I. Theory.
          J. Neurophysiology 83, pp. 808–827.
  """

  def __init__(
      self,
      size: Shape,
      V_rest: Union[float, Tensor, Initializer, Callable] = -65.,
      V_reset: Union[float, Tensor, Initializer, Callable] = -68.,
      V_th: Union[float, Tensor, Initializer, Callable] = -30.,
      V_c: Union[float, Tensor, Initializer, Callable] = -50.0,
      c: Union[float, Tensor, Initializer, Callable] = .07,
      R: Union[float, Tensor, Initializer, Callable] = 1.,
      tau: Union[float, Tensor, Initializer, Callable] = 10.,
      tau_ref: Union[float, Tensor, Initializer, Callable] = 0.,
      V_initializer: Union[Initializer, Callable, Tensor] = ZeroInit(),
      method: str = 'exp_auto',
      name: str = None
  ):
    # initialization
    super(QuaIF, self).__init__(size=size, name=name)

    # parameters
    self.V_rest = init_param(V_rest, self.num, allow_none=False)
    self.V_reset = init_param(V_reset, self.num, allow_none=False)
    self.V_th = init_param(V_th, self.num, allow_none=False)
    self.V_c = init_param(V_c, self.num, allow_none=False)
    self.c = init_param(c, self.num, allow_none=False)
    self.R = init_param(R, self.num, allow_none=False)
    self.tau = init_param(tau, self.num, allow_none=False)
    self.tau_ref = init_param(tau_ref, self.num, allow_none=False)

    # variables
    self.V = bm.Variable(init_param(V_initializer, (self.num,)))
    self.input = bm.Variable(bm.zeros(self.num))
    self.spike = bm.Variable(bm.zeros(self.num, dtype=bool))
    self.refractory = bm.Variable(bm.zeros(self.num, dtype=bool))
    self.t_last_spike = bm.Variable(bm.ones(self.num) * -1e7)

    # integral
    self.integral = odeint(method=method, f=self.derivative)

  def derivative(self, V, t, I_ext):
    dVdt = (self.c * (V - self.V_rest) * (V - self.V_c) + self.R * I_ext) / self.tau
    return dVdt

  def update(self, _t, _dt, **kwargs):
    refractory = (_t - self.t_last_spike) <= self.tau_ref
    V = self.integral(self.V, _t, self.input, dt=_dt)
    V = bm.where(refractory, self.V, V)
    spike = self.V_th <= V
    self.t_last_spike.value = bm.where(spike, _t, self.t_last_spike)
    self.V.value = bm.where(spike, self.V_reset, V)
    self.refractory.value = bm.logical_or(refractory, spike)
    self.spike.value = spike
    self.input[:] = 0.


class AdQuaIF(NeuGroup):
  r"""Adaptive quadratic integrate-and-fire neuron model.

  **Model Descriptions**

  The adaptive quadratic integrate-and-fire neuron model [1]_ is given by:

  .. math::

      \begin{aligned}
      \tau_m \frac{d V}{d t}&=c(V-V_{rest})(V-V_c) - w + I(t), \\
      \tau_w \frac{d w}{d t}&=a(V-V_{rest}) - w,
      \end{aligned}

  once the membrane potential reaches the spike threshold,

  .. math::

      V \rightarrow V_{reset}, \\
      w \rightarrow w+b.

  **Model Examples**

  .. plot::
    :include-source: True

    >>> import brainpy as bp
    >>> group = bp.dyn.AdQuaIF(1, )
    >>> runner = bp.dyn.DSRunner(group, monitors=['V', 'w'], inputs=('input', 30.))
    >>> runner.run(300)
    >>> fig, gs = bp.visualize.get_figure(2, 1, 3, 8)
    >>> fig.add_subplot(gs[0, 0])
    >>> bp.visualize.line_plot(runner.mon.ts, runner.mon.V, ylabel='V')
    >>> fig.add_subplot(gs[1, 0])
    >>> bp.visualize.line_plot(runner.mon.ts, runner.mon.w, ylabel='w', show=True)

  **Model Parameters**

  ============= ============== ======== =======================================================
  **Parameter** **Init Value** **Unit** **Explanation**
  ------------- -------------- -------- -------------------------------------------------------
  V_rest         -65            mV       Resting potential.
  V_reset        -68            mV       Reset potential after spike.
  V_th           -30            mV       Threshold potential of spike and reset.
  V_c            -50            mV       Critical voltage for spike initiation. Must be larger
                                         than :math:`V_{rest}`.
  a               1              \       The sensitivity of the recovery variable :math:`u` to
                                         the sub-threshold fluctuations of the membrane
                                         potential :math:`v`
  b              .1             \        The increment of :math:`w` produced by a spike.
  c              .07             \       Coefficient describes membrane potential update.
                                         Larger than 0.
  tau            10             ms       Membrane time constant.
  tau_w          10             ms       Time constant of the adaptation current.
  ============= ============== ======== =======================================================

  **Model Variables**

  ================== ================= ==========================================================
  **Variables name** **Initial Value** **Explanation**
  ------------------ ----------------- ----------------------------------------------------------
  V                   0                 Membrane potential.
  w                   0                 Adaptation current.
  input               0                 External and synaptic input current.
  spike               False             Flag to mark whether the neuron is spiking.
  t_last_spike        -1e7              Last spike time stamp.
  ================== ================= ==========================================================

  **References**

  .. [1] Izhikevich, E. M. (2004). Which model to use for cortical spiking
         neurons?. IEEE transactions on neural networks, 15(5), 1063-1070.
  .. [2] Touboul, Jonathan. "Bifurcation analysis of a general class of
         nonlinear integrate-and-fire neurons." SIAM Journal on Applied
         Mathematics 68, no. 4 (2008): 1045-1079.
  """

  def __init__(
      self,
      size: Shape,
      V_rest: Union[float, Tensor, Initializer, Callable] = -65.,
      V_reset: Union[float, Tensor, Initializer, Callable] = -68.,
      V_th: Union[float, Tensor, Initializer, Callable] = -30.,
      V_c: Union[float, Tensor, Initializer, Callable] = -50.0,
      a: Union[float, Tensor, Initializer, Callable] = 1.,
      b: Union[float, Tensor, Initializer, Callable] = .1,
      c: Union[float, Tensor, Initializer, Callable] = .07,
      tau: Union[float, Tensor, Initializer, Callable] = 10.,
      tau_w: Union[float, Tensor, Initializer, Callable] = 10.,
      V_initializer: Union[Initializer, Callable, Tensor] = ZeroInit(),
      w_initializer: Union[Initializer, Callable, Tensor] = ZeroInit(),
      method: str = 'exp_auto',
      name: str = None
  ):
    super(AdQuaIF, self).__init__(size=size, name=name)

    # parameters
    self.V_rest = init_param(V_rest, self.num, allow_none=False)
    self.V_reset = init_param(V_reset, self.num, allow_none=False)
    self.V_th = init_param(V_th, self.num, allow_none=False)
    self.V_c = init_param(V_c, self.num, allow_none=False)
    self.c = init_param(c, self.num, allow_none=False)
    self.a = init_param(a, self.num, allow_none=False)
    self.b = init_param(b, self.num, allow_none=False)
    self.tau = init_param(tau, self.num, allow_none=False)
    self.tau_w = init_param(tau_w, self.num, allow_none=False)

    # variables
    check_initializer(V_initializer, 'V_initializer')
    check_initializer(w_initializer, 'w_initializer')
    self.V = bm.Variable(init_param(V_initializer, (self.num,)))
    self.w = bm.Variable(init_param(w_initializer, (self.num,)))
    self.input = bm.Variable(bm.zeros(self.num))
    self.spike = bm.Variable(bm.zeros(self.num, dtype=bool))
    self.t_last_spike = bm.Variable(bm.ones(self.num) * -1e7)
    self.refractory = bm.Variable(bm.zeros(self.num, dtype=bool))

    # integral
    self.integral = odeint(method=method, f=self.derivative)

  def dV(self, V, t, w, I_ext):
    dVdt = (self.c * (V - self.V_rest) * (V - self.V_c) - w + I_ext) / self.tau
    return dVdt

  def dw(self, w, t, V):
    dwdt = (self.a * (V - self.V_rest) - w) / self.tau_w
    return dwdt

  @property
  def derivative(self):
    return JointEq([self.dV, self.dw])

  def update(self, _t, _dt):
    V, w = self.integral(self.V, self.w, _t, self.input, dt=_dt)
    spike = self.V_th <= V
    self.t_last_spike.value = bm.where(spike, _t, self.t_last_spike)
    self.V.value = bm.where(spike, self.V_reset, V)
    self.w.value = bm.where(spike, w + self.b, w)
    self.spike.value = spike
    self.input[:] = 0.


class GIF(NeuGroup):
  r"""Generalized Integrate-and-Fire model.

  **Model Descriptions**

  The generalized integrate-and-fire model [1]_ is given by

  .. math::

      &\frac{d I_j}{d t} = - k_j I_j

      &\frac{d V}{d t} = ( - (V - V_{rest}) + R\sum_{j}I_j + RI) / \tau

      &\frac{d V_{th}}{d t} = a(V - V_{rest}) - b(V_{th} - V_{th\infty})

  When :math:`V` meet :math:`V_{th}`, Generalized IF neuron fires:

  .. math::

      &I_j \leftarrow R_j I_j + A_j

      &V \leftarrow V_{reset}

      &V_{th} \leftarrow max(V_{th_{reset}}, V_{th})

  Note that :math:`I_j` refers to arbitrary number of internal currents.

  **Model Examples**

  - `Detailed examples to reproduce different firing patterns <https://brainpy-examples.readthedocs.io/en/latest/neurons/Niebur_2009_GIF.html>`_

  **Model Parameters**

  ============= ============== ======== ====================================================================
  **Parameter** **Init Value** **Unit** **Explanation**
  ------------- -------------- -------- --------------------------------------------------------------------
  V_rest        -70            mV       Resting potential.
  V_reset       -70            mV       Reset potential after spike.
  V_th_inf      -50            mV       Target value of threshold potential :math:`V_{th}` updating.
  V_th_reset    -60            mV       Free parameter, should be larger than :math:`V_{reset}`.
  R             20             \        Membrane resistance.
  tau           20             ms       Membrane time constant. Compute by :math:`R * C`.
  a             0              \        Coefficient describes the dependence of
                                        :math:`V_{th}` on membrane potential.
  b             0.01           \        Coefficient describes :math:`V_{th}` update.
  k1            0.2            \        Constant pf :math:`I1`.
  k2            0.02           \        Constant of :math:`I2`.
  R1            0              \        Free parameter.
                                        Describes dependence of :math:`I_1` reset value on
                                        :math:`I_1` value before spiking.
  R2            1              \        Free parameter.
                                        Describes dependence of :math:`I_2` reset value on
                                        :math:`I_2` value before spiking.
  A1            0              \        Free parameter.
  A2            0              \        Free parameter.
  ============= ============== ======== ====================================================================

  **Model Variables**

  ================== ================= =========================================================
  **Variables name** **Initial Value** **Explanation**
  ------------------ ----------------- ---------------------------------------------------------
  V                  -70               Membrane potential.
  input              0                 External and synaptic input current.
  spike              False             Flag to mark whether the neuron is spiking.
  V_th               -50               Spiking threshold potential.
  I1                 0                 Internal current 1.
  I2                 0                 Internal current 2.
  t_last_spike       -1e7              Last spike time stamp.
  ================== ================= =========================================================

  **References**

  .. [1] Mihalaş, Ştefan, and Ernst Niebur. "A generalized linear
         integrate-and-fire neural model produces diverse spiking
         behaviors." Neural computation 21.3 (2009): 704-718.
  .. [2] Teeter, Corinne, Ramakrishnan Iyer, Vilas Menon, Nathan
         Gouwens, David Feng, Jim Berg, Aaron Szafer et al. "Generalized
         leaky integrate-and-fire models classify multiple neuron types."
         Nature communications 9, no. 1 (2018): 1-15.
  """

  def __init__(
      self,
      size: Shape,
      V_rest: Union[float, Tensor, Initializer, Callable] = -70.,
      V_reset: Union[float, Tensor, Initializer, Callable] = -70.,
      V_th_inf: Union[float, Tensor, Initializer, Callable] = -50.,
      V_th_reset: Union[float, Tensor, Initializer, Callable] = -60.,
      R: Union[float, Tensor, Initializer, Callable] = 20.,
      tau: Union[float, Tensor, Initializer, Callable] = 20.,
      a: Union[float, Tensor, Initializer, Callable] = 0.,
      b: Union[float, Tensor, Initializer, Callable] = 0.01,
      k1: Union[float, Tensor, Initializer, Callable] = 0.2,
      k2: Union[float, Tensor, Initializer, Callable] = 0.02,
      R1: Union[float, Tensor, Initializer, Callable] = 0.,
      R2: Union[float, Tensor, Initializer, Callable] = 1.,
      A1: Union[float, Tensor, Initializer, Callable] = 0.,
      A2: Union[float, Tensor, Initializer, Callable] = 0.,
      V_initializer: Union[Initializer, Callable, Tensor] = OneInit(-70.),
      I1_initializer: Union[Initializer, Callable, Tensor] = ZeroInit(),
      I2_initializer: Union[Initializer, Callable, Tensor] = ZeroInit(),
      Vth_initializer: Union[Initializer, Callable, Tensor] = OneInit(-50.),
      method: str = 'exp_auto',
      name: str = None
  ):
    # initialization
    super(GIF, self).__init__(size=size, name=name)

    # params
    self.V_rest = init_param(V_rest, self.num, allow_none=False)
    self.V_reset = init_param(V_reset, self.num, allow_none=False)
    self.V_th_inf = init_param(V_th_inf, self.num, allow_none=False)
    self.V_th_reset = init_param(V_th_reset, self.num, allow_none=False)
    self.R = init_param(R, self.num, allow_none=False)
    self.tau = init_param(tau, self.num, allow_none=False)
    self.a = init_param(a, self.num, allow_none=False)
    self.b = init_param(b, self.num, allow_none=False)
    self.k1 = init_param(k1, self.num, allow_none=False)
    self.k2 = init_param(k2, self.num, allow_none=False)
    self.R1 = init_param(R1, self.num, allow_none=False)
    self.R2 = init_param(R2, self.num, allow_none=False)
    self.A1 = init_param(A1, self.num, allow_none=False)
    self.A2 = init_param(A2, self.num, allow_none=False)

    # variables
    check_initializer(V_initializer, 'V_initializer')
    check_initializer(I1_initializer, 'I1_initializer')
    check_initializer(I2_initializer, 'I2_initializer')
    check_initializer(Vth_initializer, 'Vth_initializer')
    self.I1 = bm.Variable(init_param(I1_initializer, (self.num,)))
    self.I2 = bm.Variable(init_param(I2_initializer, (self.num,)))
    self.V = bm.Variable(init_param(V_initializer, (self.num,)))
    self.V_th = bm.Variable(init_param(Vth_initializer, (self.num,)))
    self.input = bm.Variable(bm.zeros(self.num))
    self.spike = bm.Variable(bm.zeros(self.num, dtype=bool))
    self.t_last_spike = bm.Variable(bm.ones(self.num) * -1e7)

    # integral
    self.integral = odeint(method=method, f=self.derivative)

  def dI1(self, I1, t):
    return - self.k1 * I1

  def dI2(self, I2, t):
    return - self.k2 * I2

  def dVth(self, V_th, t, V):
    return self.a * (V - self.V_rest) - self.b * (V_th - self.V_th_inf)

  def dV(self, V, t, I1, I2, I_ext):
    return (- (V - self.V_rest) + self.R * I_ext + self.R * I1 + self.R * I2) / self.tau

  @property
  def derivative(self):
    return JointEq([self.dI1, self.dI2, self.dVth, self.dV])

  def update(self, _t, _dt):
    I1, I2, V_th, V = self.integral(self.I1, self.I2, self.V_th, self.V, _t, self.input, dt=_dt)
    spike = self.V_th <= V
    V = bm.where(spike, self.V_reset, V)
    I1 = bm.where(spike, self.R1 * I1 + self.A1, I1)
    I2 = bm.where(spike, self.R2 * I2 + self.A2, I2)
    reset_th = bm.logical_and(V_th < self.V_th_reset, spike)
    V_th = bm.where(reset_th, self.V_th_reset, V_th)
    self.spike.value = spike
    self.I1.value = I1
    self.I2.value = I2
    self.V_th.value = V_th
    self.V.value = V
    self.input[:] = 0.


class Izhikevich(NeuGroup):
  r"""The Izhikevich neuron model.

  **Model Descriptions**

  The dynamics of the Izhikevich neuron model [1]_ [2]_ is given by:

  .. math ::

      \frac{d V}{d t} &= 0.04 V^{2}+5 V+140-u+I

      \frac{d u}{d t} &=a(b V-u)

  .. math ::

      \text{if}  v \geq 30  \text{mV}, \text{then}
      \begin{cases} v \leftarrow c \\
      u \leftarrow u+d \end{cases}

  **Model Examples**

  - `Detailed examples to reproduce different firing patterns <https://brainpy-examples.readthedocs.io/en/latest/neurons/Izhikevich_2003_Izhikevich_model.html>`_

  **Model Parameters**

  ============= ============== ======== ================================================================================
  **Parameter** **Init Value** **Unit** **Explanation**
  ------------- -------------- -------- --------------------------------------------------------------------------------
  a             0.02           \        It determines the time scale of
                                        the recovery variable :math:`u`.
  b             0.2            \        It describes the sensitivity of the
                                        recovery variable :math:`u` to
                                        the sub-threshold fluctuations of the
                                        membrane potential :math:`v`.
  c             -65            \        It describes the after-spike reset value
                                        of the membrane potential :math:`v` caused by
                                        the fast high-threshold :math:`K^{+}`
                                        conductance.
  d             8              \        It describes after-spike reset of the
                                        recovery variable :math:`u`
                                        caused by slow high-threshold
                                        :math:`Na^{+}` and :math:`K^{+}` conductance.
  tau_ref       0              ms       Refractory period length. [ms]
  V_th          30             mV       The membrane potential threshold.
  ============= ============== ======== ================================================================================

  **Model Variables**

  ================== ================= =========================================================
  **Variables name** **Initial Value** **Explanation**
  ------------------ ----------------- ---------------------------------------------------------
  V                          -65        Membrane potential.
  u                          1          Recovery variable.
  input                      0          External and synaptic input current.
  spike                      False      Flag to mark whether the neuron is spiking.
  refractory                False       Flag to mark whether the neuron is in refractory period.
  t_last_spike               -1e7       Last spike time stamp.
  ================== ================= =========================================================

  **References**

  .. [1] Izhikevich, Eugene M. "Simple model of spiking neurons." IEEE
         Transactions on neural networks 14.6 (2003): 1569-1572.

  .. [2] Izhikevich, Eugene M. "Which model to use for cortical spiking neurons?."
         IEEE transactions on neural networks 15.5 (2004): 1063-1070.
  """

  def __init__(
      self,
      size: Shape,
      a: Union[float, Tensor, Initializer, Callable] = 0.02,
      b: Union[float, Tensor, Initializer, Callable] = 0.20,
      c: Union[float, Tensor, Initializer, Callable] = -65.,
      d: Union[float, Tensor, Initializer, Callable] = 8.,
      V_th: Union[float, Tensor, Initializer, Callable] = 30.,
      tau_ref: Union[float, Tensor, Initializer, Callable] = 0.,
      V_initializer: Union[Initializer, Callable, Tensor] = ZeroInit(),
      u_initializer: Union[Initializer, Callable, Tensor] = OneInit(),
      method: str = 'exp_auto',
      name: str = None
  ):
    # initialization
    super(Izhikevich, self).__init__(size=size, name=name)

    # params
    self.a = init_param(a, self.num, allow_none=False)
    self.b = init_param(b, self.num, allow_none=False)
    self.c = init_param(c, self.num, allow_none=False)
    self.d = init_param(d, self.num, allow_none=False)
    self.V_th = init_param(V_th, self.num, allow_none=False)
    self.tau_ref = init_param(tau_ref, self.num, allow_none=False)

    # variables
    check_initializer(V_initializer, 'V_initializer')
    check_initializer(u_initializer, 'u_initializer')
    self.u = bm.Variable(init_param(u_initializer, (self.num,)))
    self.V = bm.Variable(init_param(V_initializer, (self.num,)))
    self.input = bm.Variable(bm.zeros(self.num))
    self.spike = bm.Variable(bm.zeros(self.num, dtype=bool))
    self.refractory = bm.Variable(bm.zeros(self.num, dtype=bool))
    self.t_last_spike = bm.Variable(bm.ones(self.num) * -1e7)

    # functions
    self.integral = odeint(method=method, f=JointEq([self.dV, self.du]))

  def dV(self, V, t, u, I_ext):
    dVdt = 0.04 * V * V + 5 * V + 140 - u + I_ext
    return dVdt

  def du(self, u, t, V):
    dudt = self.a * (self.b * V - u)
    return dudt

  def update(self, _t, _dt):
    V, u = self.integral(self.V, self.u, _t, self.input, dt=_dt)
    refractory = (_t - self.t_last_spike) <= self.tau_ref
    V = bm.where(refractory, self.V, V)
    spike = self.V_th <= V
    self.t_last_spike.value = bm.where(spike, _t, self.t_last_spike)
    self.V.value = bm.where(spike, self.c, V)
    self.u.value = bm.where(spike, u + self.d, u)
    self.refractory.value = bm.logical_or(refractory, spike)
    self.spike.value = spike
    self.input[:] = 0.


class HindmarshRose(NeuGroup):
  r"""Hindmarsh-Rose neuron model.

  **Model Descriptions**

  The Hindmarsh–Rose model [1]_ [2]_ of neuronal activity is aimed to study the
  spiking-bursting behavior of the membrane potential observed in experiments
  made with a single neuron.

  The model has the mathematical form of a system of three nonlinear ordinary
  differential equations on the dimensionless dynamical variables :math:`x(t)`,
  :math:`y(t)`, and :math:`z(t)`. They read:

  .. math::

     \begin{aligned}
     \frac{d V}{d t} &= y - a V^3 + b V^2 - z + I \\
     \frac{d y}{d t} &= c - d V^2 - y \\
     \frac{d z}{d t} &= r (s (V - V_{rest}) - z)
     \end{aligned}

  where :math:`a, b, c, d` model the working of the fast ion channels,
  :math:`I` models the slow ion channels.

  **Model Examples**

  .. plot::
    :include-source: True

    >>> import brainpy as bp
    >>> import matplotlib.pyplot as plt
    >>>
    >>> bp.math.set_dt(dt=0.01)
    >>> bp.ode.set_default_odeint('rk4')
    >>>
    >>> types = ['quiescence', 'spiking', 'bursting', 'irregular_spiking', 'irregular_bursting']
    >>> bs = bp.math.array([1.0, 3.5, 2.5, 2.95, 2.8])
    >>> Is = bp.math.array([2.0, 5.0, 3.0, 3.3, 3.7])
    >>>
    >>> # define neuron type
    >>> group = bp.dyn.HindmarshRose(len(types), b=bs)
    >>> runner = bp.dyn.DSRunner(group, monitors=['V'], inputs=['input', Is],)
    >>> runner.run(1e3)
    >>>
    >>> fig, gs = bp.visualize.get_figure(row_num=3, col_num=2, row_len=3, col_len=5)
    >>> for i, mode in enumerate(types):
    >>>     fig.add_subplot(gs[i // 2, i % 2])
    >>>     plt.plot(runner.mon.ts, runner.mon.V[:, i])
    >>>     plt.title(mode)
    >>>     plt.xlabel('Time [ms]')
    >>> plt.show()

  **Model Parameters**

  ============= ============== ========= ============================================================
  **Parameter** **Init Value** **Unit**  **Explanation**
  ------------- -------------- --------- ------------------------------------------------------------
  a             1              \         Model parameter.
                                         Fixed to a value best fit neuron activity.
  b             3              \         Model parameter.
                                         Allows the model to switch between bursting
                                         and spiking, controls the spiking frequency.
  c             1              \         Model parameter.
                                         Fixed to a value best fit neuron activity.
  d             5              \         Model parameter.
                                         Fixed to a value best fit neuron activity.
  r             0.01           \         Model parameter.
                                         Controls slow variable z's variation speed.
                                         Governs spiking frequency when spiking, and affects the
                                         number of spikes per burst when bursting.
  s             4              \         Model parameter. Governs adaption.
  ============= ============== ========= ============================================================

  **Model Variables**

  =============== ================= =====================================
  **Member name** **Initial Value** **Explanation**
  --------------- ----------------- -------------------------------------
  V               -1.6              Membrane potential.
  y               -10               Gating variable.
  z               0                 Gating variable.
  spike           False             Whether generate the spikes.
  input           0                 External and synaptic input current.
  t_last_spike    -1e7              Last spike time stamp.
  =============== ================= =====================================

  **References**

  .. [1] Hindmarsh, James L., and R. M. Rose. "A model of neuronal bursting using
        three coupled first order differential equations." Proceedings of the
        Royal society of London. Series B. Biological sciences 221.1222 (1984):
        87-102.
  .. [2] Storace, Marco, Daniele Linaro, and Enno de Lange. "The Hindmarsh–Rose
        neuron model: bifurcation analysis and piecewise-linear approximations."
        Chaos: An Interdisciplinary Journal of Nonlinear Science 18.3 (2008):
        033128.
  """

  def __init__(
      self,
      size: Shape,
      a: Union[float, Tensor, Initializer, Callable] = 1.,
      b: Union[float, Tensor, Initializer, Callable] = 3.,
      c: Union[float, Tensor, Initializer, Callable] = 1.,
      d: Union[float, Tensor, Initializer, Callable] = 5.,
      r: Union[float, Tensor, Initializer, Callable] = 0.01,
      s: Union[float, Tensor, Initializer, Callable] = 4.,
      V_rest: Union[float, Tensor, Initializer, Callable] = -1.6,
      V_th: Union[float, Tensor, Initializer, Callable] = 1.0,
      V_initializer: Union[Initializer, Callable, Tensor] = ZeroInit(),
      y_initializer: Union[Initializer, Callable, Tensor] = OneInit(-10.),
      z_initializer: Union[Initializer, Callable, Tensor] = ZeroInit(),
      method: str = 'exp_auto',
      name: str = None
  ):
    # initialization
    super(HindmarshRose, self).__init__(size=size, name=name)

    # parameters
    self.a = init_param(a, self.num, allow_none=False)
    self.b = init_param(b, self.num, allow_none=False)
    self.c = init_param(c, self.num, allow_none=False)
    self.d = init_param(d, self.num, allow_none=False)
    self.r = init_param(r, self.num, allow_none=False)
    self.s = init_param(s, self.num, allow_none=False)
    self.V_th = init_param(V_th, self.num, allow_none=False)
    self.V_rest = init_param(V_rest, self.num, allow_none=False)

    # variables
    check_initializer(V_initializer, 'V_initializer')
    check_initializer(y_initializer, 'y_initializer')
    check_initializer(z_initializer, 'z_initializer')
    self.z = bm.Variable(init_param(V_initializer, (self.num,)))
    self.y = bm.Variable(init_param(y_initializer, (self.num,)))
    self.V = bm.Variable(init_param(z_initializer, (self.num,)))
    self.input = bm.Variable(bm.zeros(self.num))
    self.spike = bm.Variable(bm.zeros(self.num, dtype=bool))
    self.t_last_spike = bm.Variable(bm.ones(self.num) * -1e7)

    # integral
    self.integral = odeint(method=method, f=self.derivative)

  def dV(self, V, t, y, z, I_ext):
    return y - self.a * V * V * V + self.b * V * V - z + I_ext

  def dy(self, y, t, V):
    return self.c - self.d * V * V - y

  def dz(self, z, t, V):
    return self.r * (self.s * (V - self.V_rest) - z)

  @property
  def derivative(self):
    return JointEq([self.dV, self.dy, self.dz])

  def update(self, _t, _dt):
    V, y, z = self.integral(self.V, self.y, self.z, _t, self.input, dt=_dt)
    self.spike.value = bm.logical_and(V >= self.V_th, self.V < self.V_th)
    self.t_last_spike.value = bm.where(self.spike, _t, self.t_last_spike)
    self.V.value = V
    self.y.value = y
    self.z.value = z
    self.input[:] = 0.


class FHN(NeuGroup):
  r"""FitzHugh-Nagumo neuron model.

  **Model Descriptions**

  The FitzHugh–Nagumo model (FHN), named after Richard FitzHugh (1922–2007)
  who suggested the system in 1961 [1]_ and J. Nagumo et al. who created the
  equivalent circuit the following year, describes a prototype of an excitable
  system (e.g., a neuron).

  The motivation for the FitzHugh-Nagumo model was to isolate conceptually
  the essentially mathematical properties of excitation and propagation from
  the electrochemical properties of sodium and potassium ion flow. The model
  consists of

  - a *voltage-like variable* having cubic nonlinearity that allows regenerative
    self-excitation via a positive feedback, and
  - a *recovery variable* having a linear dynamics that provides a slower negative feedback.

  .. math::

     \begin{aligned}
     {\dot {v}} &=v-{\frac {v^{3}}{3}}-w+RI_{\rm {ext}},  \\
     \tau {\dot  {w}}&=v+a-bw.
     \end{aligned}

  The FHN Model is an example of a relaxation oscillator
  because, if the external stimulus :math:`I_{\text{ext}}`
  exceeds a certain threshold value, the system will exhibit
  a characteristic excursion in phase space, before the
  variables :math:`v` and :math:`w` relax back to their rest values.
  This behaviour is typical for spike generations (a short,
  nonlinear elevation of membrane voltage :math:`v`,
  diminished over time by a slower, linear recovery variable
  :math:`w`) in a neuron after stimulation by an external
  input current.

  **Model Examples**

  .. plot::
    :include-source: True

    >>> import brainpy as bp
    >>> fhn = bp.dyn.FHN(1)
    >>> runner = bp.dyn.DSRunner(fhn, inputs=('input', 1.), monitors=['V', 'w'])
    >>> runner.run(100.)
    >>> bp.visualize.line_plot(runner.mon.ts, runner.mon.w, legend='w')
    >>> bp.visualize.line_plot(runner.mon.ts, runner.mon.V, legend='V', show=True)

  **Model Parameters**

  ============= ============== ======== ========================
  **Parameter** **Init Value** **Unit** **Explanation**
  ------------- -------------- -------- ------------------------
  a             1              \        Positive constant
  b             1              \        Positive constant
  tau           10             ms       Membrane time constant.
  V_th          1.8            mV       Threshold potential of spike.
  ============= ============== ======== ========================

  **Model Variables**

  ================== ================= =========================================================
  **Variables name** **Initial Value** **Explanation**
  ------------------ ----------------- ---------------------------------------------------------
  V                   0                 Membrane potential.
  w                   0                 A recovery variable which represents
                                        the combined effects of sodium channel
                                        de-inactivation and potassium channel
                                        deactivation.
  input               0                 External and synaptic input current.
  spike               False             Flag to mark whether the neuron is spiking.
  t_last_spike       -1e7               Last spike time stamp.
  ================== ================= =========================================================

  **References**

  .. [1] FitzHugh, Richard. "Impulses and physiological states in theoretical models of nerve membrane." Biophysical journal 1.6 (1961): 445-466.
  .. [2] https://en.wikipedia.org/wiki/FitzHugh%E2%80%93Nagumo_model
  .. [3] http://www.scholarpedia.org/article/FitzHugh-Nagumo_model

  """

  def __init__(
      self,
      size: Shape,
      a: Union[float, Tensor, Initializer, Callable] = 0.7,
      b: Union[float, Tensor, Initializer, Callable] = 0.8,
      tau: Union[float, Tensor, Initializer, Callable] = 12.5,
      Vth: Union[float, Tensor, Initializer, Callable] = 1.8,
      V_initializer: Union[Initializer, Callable, Tensor] = ZeroInit(),
      w_initializer: Union[Initializer, Callable, Tensor] = ZeroInit(),
      method: str = 'exp_auto',
      name: str = None
  ):
    # initialization
    super(FHN, self).__init__(size=size, name=name)

    # parameters
    self.a = init_param(a, self.num, allow_none=False)
    self.b = init_param(b, self.num, allow_none=False)
    self.tau = init_param(tau, self.num, allow_none=False)
    self.Vth = init_param(Vth, self.num, allow_none=False)

    # variables
    check_initializer(V_initializer, 'V_initializer')
    check_initializer(w_initializer, 'w_initializer')
    self.w = bm.Variable(init_param(w_initializer, (self.num,)))
    self.V = bm.Variable(init_param(V_initializer, (self.num,)))
    self.input = bm.Variable(bm.zeros(self.num))
    self.spike = bm.Variable(bm.zeros(self.num, dtype=bool))
    self.t_last_spike = bm.Variable(bm.ones(self.num) * -1e7)

    # integral
    self.integral = odeint(method=method, f=self.derivative)

  def dV(self, V, t, w, I_ext):
    return V - V * V * V / 3 - w + I_ext

  def dw(self, w, t, V):
    return (V + self.a - self.b * w) / self.tau

  @property
  def derivative(self):
    return JointEq([self.dV, self.dw])

  def update(self, _t, _dt):
    V, w = self.integral(self.V, self.w, _t, self.input, dt=_dt)
    self.spike.value = bm.logical_and(V >= self.Vth, self.V < self.Vth)
    self.t_last_spike.value = bm.where(self.spike, _t, self.t_last_spike)
    self.V.value = V
    self.w.value = w
    self.input[:] = 0.



# -*- coding: utf-8 -*-

from functools import partial

import numpy as np
from jax import vmap
from jax import numpy as jnp
from jax.scipy.optimize import minimize

import brainpy.math as bm
from brainpy import errors, tools
from brainpy.analysis import constants as C, utils
from brainpy.base.collector import Collector

pyplot = None

__all__ = [
  'LowDimAnalyzer',
  'Num1DAnalyzer',
  'Num2DAnalyzer',
]


class LowDimAnalyzer(object):
  r"""Automatic Analyzer for Low-dimensional Dynamical Systems.

  A dynamical model is characterized by a series of dynamical
  variables and parameters:

  .. math::

      {dF \over dt} = F(v_1, v_2, ..., p_1, p_2, ...)

  where :math:`v_1, v_2` are variables, :math:`p_1, p_2` are parameters.

  .. note::
    ``LowDimAnalyzer`` cannot analyze dynamical system depends on time :math:`t`.

  Parameters
  ----------
  model : Any, ODEIntegrator, sequence of ODEIntegrator, DynamicalSystem
      A model of the population, the integrator function,
      or a list/tuple of integrator functions.
  target_vars : dict
      The target/dynamical variables.
  fixed_vars : dict
      The fixed variables.
  target_pars : dict, optional
      The parameters which can be dynamical varied.
  pars_update : dict, optional
      The parameters to update.
  resolutions : float, dict
      The resolution for numerical iterative solvers. Default is 20 equal parts
      (:math:`\frac{\mathrm{max} - \mathrm{min}}{20}`). It can
      set the numerical resolution of dynamical variables or dynamical parameters.
      For example,

      - set ``resolutions=0.1`` will generalize it to all variables and parameters;
      - set ``resolutions={var1: 0.1, var2: 0.2, par1: 0.1, par2: 0.05}`` will specify
        the particular resolutions to variables and parameters.
      - Moreover, you can also set ``resolutions={var1: JaxArray([...]), var2: 0.1}``
        to specify the search points need to explore for variable `var1`.
        This will be useful to set sense search points at some inflection points.
  lim_scale: float
    The axis limit scale factor. Default is 1.05. The setting means
    the axes will be clipped to ``[var_min * (1-lim_scale)/2, var_max * (var_max-1)/2]``.
  options : optional, dict
    The optional setting. Maybe needed in the individual analyzer.
  """

  def __init__(self,
               model,
               target_vars,
               fixed_vars=None,
               target_pars=None,
               pars_update=None,
               resolutions=None,
               jit_device=None,
               lim_scale=1.05,
               options=None, ):
    # model
    # -----
    self.model = utils.model_transform(model)

    # target variables
    # ----------------
    if not isinstance(target_vars, dict):
      raise errors.AnalyzerError('"target_vars" must be a dict, with the format of '
                                 '{"var1": (var1_min, var1_max)}.')
    self.target_vars = Collector(target_vars)
    self.target_var_names = list(self.target_vars.keys())  # list of target vars
    for key in self.target_vars.keys():
      if key not in self.model.variables:
        raise errors.AnalyzerError(f'{key} is not a dynamical variable in {self.model}.')

    # fixed variables
    # ----------------
    if fixed_vars is None:
      fixed_vars = dict()
    if not isinstance(fixed_vars, dict):
      raise errors.AnalyzerError('"fixed_vars" must be a dict with the format '
                                 'of {"var1": val1, "var2": val2}.')
    for key in fixed_vars.keys():
      if key not in self.model.variables:
        raise ValueError(f'{key} is not a dynamical variable in {self.model}.')
    self.fixed_vars = Collector(fixed_vars)

    # check duplicate
    for key in self.fixed_vars.keys():
      if key in self.target_vars:
        raise errors.AnalyzerError(f'"{key}" is defined as a target variable in "target_vars", '
                                   f'but also defined as a fixed variable in "fixed_vars".')

    # parameters to update
    # ---------------------
    if pars_update is None:
      pars_update = dict()
    if not isinstance(pars_update, dict):
      raise errors.AnalyzerError('"pars_update" must be a dict with the format '
                                 'of {"par1": val1, "par2": val2}.')
    pars_update = Collector(pars_update)
    for key in pars_update.keys():
      if key not in self.model.parameters:
        raise errors.AnalyzerError(f'"{key}" is not a valid parameter in "{self.model}" model.')
    self.pars_update = pars_update

    # dynamical parameters
    # ---------------------
    if target_pars is None:
      target_pars = dict()
    if not isinstance(target_pars, dict):
      raise errors.AnalyzerError('"target_pars" must be a dict with the format of {"par1": (val1, val2)}.')
    for key in target_pars.keys():
      if key not in self.model.parameters:
        raise errors.AnalyzerError(f'"{key}" is not a valid parameter in "{self.model}" model.')
    self.target_pars = Collector(target_pars)
    self.target_par_names = list(self.target_pars.keys())  # list of target_pars

    # check duplicate
    for key in self.pars_update.keys():
      if key in self.target_pars:
        raise errors.AnalyzerError(f'"{key}" is defined as a target parameter in "target_pars", '
                                   f'but also defined as a fixed parameter in "pars_update".')

    # resolutions for numerical methods
    # ---------------------------------
    self.resolutions = dict()
    _target_vp = self.target_vars + self.target_pars
    if resolutions is None:
      for key, lim in self.target_vars.items():
        self.resolutions[key] = bm.asarray(np.linspace(*lim, 20))
      for key, lim in self.target_pars.items():
        self.resolutions[key] = bm.asarray(np.linspace(*lim, 20))
    elif isinstance(resolutions, float):
      for key, lim in self.target_vars.items():
        self.resolutions[key] = bm.asarray(np.arange(*lim, resolutions))
      for key, lim in self.target_pars.items():
        self.resolutions[key] = bm.asarray(np.arange(*lim, resolutions))
    elif isinstance(resolutions, dict):
      for key in resolutions.keys():
        if key in self.target_var_names:
          continue
        if key in self.target_par_names:
          continue
        raise errors.AnalyzerError(f'The resolution setting target "{key}" is not found in '
                                   f'the target variables {self.target_var_names} and '
                                   f'the target parameters {self.target_par_names}.')
      for key in self.target_var_names + self.target_par_names:
        if key not in resolutions:
          self.resolutions[key] = bm.asarray(np.linspace(*_target_vp[key], 20))
        else:
          resolution = resolutions[key]
          if isinstance(resolution, float):
            self.resolutions[key] = bm.asarray(np.arange(*_target_vp[key], resolution))
          elif isinstance(resolution, (bm.ndarray, np.ndarray, jnp.ndarray)):
            if not np.ndim(resolution) == 1:
              raise errors.AnalyzerError(f'resolution must be a 1D array, but get its '
                                         f'shape with {resolution.shape}.')
            self.resolutions[key] = bm.asarray(resolution)
          else:
            raise errors.AnalyzerError(f'Unknown resolution setting: {key}: {resolution}')
    else:
      raise errors.AnalyzerError(f'Unknown resolution type: {type(resolutions)}')

    # other settings
    # --------------
    if options is None:
      options = dict()
    self.options = options
    self.jit_device = jit_device
    self.lim_scale = lim_scale

    # A dict to store the analyzed results
    # -------------------------------------
    # 'dxdt' : The differential function ``f`` of the first variable ``x``.
    #          It can be used as ``dxdt(x, y, ...)``.
    # 'dydt' : The differential function ``g`` of the second variable ``y``.
    #          It can be used as ``dydt(x, y, ...)``.
    # 'dfdx' : The derivative of ``f`` by ``x``. It can be used as ``dfdx(x, y, ...)``.
    # 'dfdy' : The derivative of ``f`` by ``y``. It can be used as ``dfdy(x, y, ...)``.
    # 'dgdx' : The derivative of ``g`` by ``x``. It can be used as ``dgdx(x, y, ...)``.
    # 'dgdy' : The derivative of ``g`` by ``y``. It can be used as ``dgdy(x, y, ...)``.
    # 'jacobian' : The jacobian matrix. It can be used as ``jacobian(x, y, ...)``.
    # 'fixed_point' : The fixed point.
    # 'y_by_x_in_fy' :
    # 'x_by_y_in_fy' :
    # 'y_by_x_in_fx' :
    # 'x_by_y_in_fx' :
    self.analyzed_results = tools.DictPlus()

  def show_figure(self):
    global pyplot
    if pyplot is None:
      from matplotlib import pyplot
    pyplot.show()


class Num1DAnalyzer(LowDimAnalyzer):
  r"""Analyzer for one-dimensional dynamical system.

  It supports the analysis for 1D dynamical system.

  .. math::

      {dx \over dt} = f(x, t)

  Actually, the analysis for 1D system is purely numerically.
  """

  def __init__(self, *args, **kwargs):
    super(Num1DAnalyzer, self).__init__(*args, **kwargs)
    self.x_var = self.target_var_names[0]
    if len(self.target_vars) < 1:
      raise errors.AnalyzerError(f'{Num1DAnalyzer.__name__} only supports dynamical system '
                                 f'with >= 1 variables. But we got {len(self.target_vars)} '
                                 f'variables in {self.model}.')

  @property
  def F_fx(self):
    """Make the standard function call of :math:`f_x (*\mathrm{vars}, *\mathrm{pars})`.

    This function has been transformed into the standard call.
    For instance, if the user has the ``target_vars=("v1", "v2")`` and
    the ``target_pars=("p1", "p2")``, while the first function is defined as:

    >>> def f1(v1, t, p1):
    >>>   return something

    However, after the stransformation, this function should be called as:

    >>> self.F_fx(v1, v2, p1, p2)
    """
    if C.F_fx not in self.analyzed_results:
      _, arguments = utils.get_args(self.model.F[self.x_var])
      wrapper = utils.std_derivative(arguments, self.target_var_names, self.target_par_names)
      f = wrapper(self.model.F[self.x_var])
      f = partial(f, **(self.pars_update + self.fixed_vars))
      f = utils.f_without_jaxarray_return(f)
      f = utils.remove_return_shape(f)
      self.analyzed_results[C.F_fx] = bm.jit(f, device=self.jit_device)
    return self.analyzed_results[C.F_fx]

  @property
  def F_vmap_fx(self):
    if C.F_vmap_fx not in self.analyzed_results:
      self.analyzed_results[C.F_vmap_fx] = bm.jit(vmap(self.F_fx), device=self.jit_device)
    return self.analyzed_results[C.F_vmap_fx]

  @property
  def F_dfxdx(self):
    """The function to evaluate :math:`\frac{df_x(*\mathrm{vars}, *\mathrm{pars})}{dx}`."""
    if C.F_dfxdx not in self.analyzed_results:
      dfx = bm.vector_grad(self.F_fx, argnums=0)
      self.analyzed_results[C.F_dfxdx] = bm.jit(dfx, device=self.jit_device)
    return self.analyzed_results[C.F_dfxdx]

  @property
  def F_fixed_point_aux(self):
    if C.F_fixed_point_aux not in self.analyzed_results:
      def aux_fun(x, *args):
        return jnp.abs(self.F_fx(x, *args)).sum()

      self.analyzed_results[C.F_fixed_point_aux] = aux_fun
    return self.analyzed_results[C.F_fixed_point_aux]

  @property
  def F_vmap_fp_aux(self):
    if C.F_vmap_fp_aux not in self.analyzed_results:
      # The arguments of this function are:
      # ---
      # "X": a two-dimensional matrix: (num_batch, num_var)
      # "args": a list of one-dimensional vectors, each has the shape of (num_batch,)
      self.analyzed_results[C.F_vmap_fp_aux] = bm.jit(vmap(self.F_fixed_point_aux))
    return self.analyzed_results[C.F_vmap_fp_aux]

  @property
  def F_fixed_point_opt(self):
    if C.F_fixed_point_opt not in self.analyzed_results:
      def f(start_and_end, *args):
        return utils.jax_brentq(self.F_fx)(start_and_end[0], start_and_end[1], args)

      self.analyzed_results[C.F_fixed_point_opt] = f
    return self.analyzed_results[C.F_fixed_point_opt]

  @property
  def F_vmap_fp_opt(self):
    if C.F_vmap_fp_opt not in self.analyzed_results:
      # The arguments of this function are:
      # ---
      # "X": a two-dimensional matrix: (num_batch, num_var)
      # "args": a list of one-dimensional vectors, each has the shape of (num_batch,)
      self.analyzed_results[C.F_vmap_fp_opt] = bm.jit(vmap(self.F_fixed_point_opt))
    return self.analyzed_results[C.F_vmap_fp_opt]

  def _get_fixed_points(self, candidates, *args, num_seg=None, tol_aux=1e-7, loss_screen=None):
    """

    "candidates" and "args" can be obtained through:

    >>> all_candidates = []
    >>> all_par1 = []
    >>> all_par2 = []
    >>> for p1 in par1_list:
    >>>   for p2 in par2_list:
    >>>     xs = self.resolutions[self.x_var]
    >>>     all_candidates.append(xs)
    >>>     all_par1.append(jnp.ones_like(xs) * p1)
    >>>     all_par2.append(jnp.ones_like(xs) * p2)

    Parameters
    ----------
    candidates
    args
    tol_aux
    loss_screen

    Returns
    -------

    """
    # candidates: xs, a vector with the length of self.resolutions[self.x_var]
    # args: parameters, a list/tuple of vectors
    candidates = candidates.value if isinstance(candidates, bm.JaxArray) else candidates
    selected_ids = np.arange(len(candidates))
    args = tuple(a.value if isinstance(candidates, bm.JaxArray) else a for a in args)
    for a in args: assert len(a) == len(candidates)
    if num_seg is None:
      num_seg = len(self.resolutions[self.x_var])
    assert isinstance(num_seg, int)

    # get the signs
    signs = jnp.sign(self.F_vmap_fx(candidates, *args))
    signs = signs.reshape((num_seg, -1))
    par_len = signs.shape[1]
    signs1 = signs.at[-1].set(1)
    signs2 = jnp.vstack((signs[1:], signs[:1])).at[-1].set(1)
    ids = jnp.where((signs1 * signs2).flatten() <= 0)[0]
    if len(ids) <= 0:
      return [], [], []

    # selected the proper candidates to optimize fixed points
    selected_ids = selected_ids[np.asarray(ids)]
    starts = candidates[ids]
    ends = candidates[ids + par_len]
    X = jnp.stack((starts, ends)).T
    args = tuple(a[ids] for a in args)

    # optimize the fixed points
    res = self.F_vmap_fp_opt(X, *args)
    losses = self.F_vmap_fp_aux(res['root'], *args)
    valid_or_not = jnp.logical_and(res['status'] == utils.ECONVERGED, losses <= tol_aux)
    ids = np.asarray(jnp.where(valid_or_not)[0])
    fps = np.asarray(res['root'])[ids]
    args = tuple(a[ids] for a in args)
    selected_ids = selected_ids[np.asarray(ids)]
    return fps, selected_ids, args


class Num2DAnalyzer(Num1DAnalyzer):
  r"""Analyzer for two-dimensional dynamical system.

  It supports the analysis for 2D dynamical system.

  .. math::

      {dx \over dt} = fx(x, t, y)

      {dy \over dt} = fy(y, t, x)
  """

  def __init__(self, *args, **kwargs):
    super(Num2DAnalyzer, self).__init__(*args, **kwargs)
    if len(self.target_vars) < 2:
      raise errors.AnalyzerError(f'{Num1DAnalyzer.__name__} only supports dynamical system '
                                 f'with >= 2 variables. But we got {len(self.target_vars)} '
                                 f'variables in {self.model}.')
    self.y_var = self.target_var_names[1]

  @property
  def F_fy(self):
    """The function to evaluate :math:`f_y(*\mathrm{vars}, *\mathrm{pars})`.

    This function has been transformed into the standard call.
    For instance, if the user has the ``target_vars=("v1", "v2")`` and
    the ``target_pars=("p1", "p2")``, while the first function is defined as:

    >>> def f1(v1, t, p1):
    >>>   return something

    However, after the stransformation, this function should be called as:

    >>> self.F_fy(v1, v2, p1, p2)
    """
    if C.F_fy not in self.analyzed_results:
      variables, arguments = utils.get_args(self.model.F[self.y_var])
      wrapper = utils.std_derivative(arguments, self.target_var_names, self.target_par_names)
      f = wrapper(self.model.F[self.y_var])
      f = partial(f, **(self.pars_update + self.fixed_vars))
      f = utils.f_without_jaxarray_return(f)
      f = utils.remove_return_shape(f)
      self.analyzed_results[C.F_fy] = bm.jit(f, device=self.jit_device)
    return self.analyzed_results[C.F_fy]

  @property
  def F_int_x(self):
    if C.F_int_x not in self.analyzed_results:
      wrap_x = utils.std_derivative(utils.get_args(self.model.F[self.x_var])[1],
                                    self.target_var_names, self.target_par_names)
      init_x = partial(wrap_x(self.model.INTG[0]), **(self.pars_update + self.fixed_vars))
      self.analyzed_results[C.F_int_x] = init_x
    return self.analyzed_results[C.F_int_x]

  @property
  def F_int_y(self):
    if C.F_int_y not in self.analyzed_results:
      wrap_x = utils.std_derivative(utils.get_args(self.model.F[self.y_var])[1],
                                    self.target_var_names, self.target_par_names)
      init_x = partial(wrap_x(self.model.INTG[1]), **(self.pars_update + self.fixed_vars))
      self.analyzed_results[C.F_int_y] = init_x
    return self.analyzed_results[C.F_int_y]

  @property
  def F_x_by_y_in_fx(self):
    if C.F_x_by_y_in_fx not in self.analyzed_results:
      if C.x_by_y_in_fx in self.options:
        wrapper = utils.std_func(utils.get_args(self.options[C.x_by_y_in_fx], gather_var=False),
                                 self.target_var_names[1:],
                                 self.target_par_names)
        f = wrapper(self.options[C.x_by_y_in_fx])
        f = partial(f, **(self.pars_update + self.fixed_vars))
        f = utils.f_without_jaxarray_return(f)
        self.analyzed_results[C.F_x_by_y_in_fx] = f
      else:
        self.analyzed_results[C.F_x_by_y_in_fx] = None
    return self.analyzed_results[C.F_x_by_y_in_fx]

  @property
  def F_y_by_x_in_fx(self):
    if C.F_y_by_x_in_fx not in self.analyzed_results:
      if C.y_by_x_in_fx in self.options:
        wrapper = utils.std_func(utils.get_args(self.options[C.y_by_x_in_fx], gather_var=False),
                                 self.target_var_names[:1] + self.target_var_names[2:],
                                 self.target_par_names)
        f = wrapper(self.options[C.y_by_x_in_fx])
        f = partial(f, **(self.pars_update + self.fixed_vars))
        f = utils.f_without_jaxarray_return(f)
        self.analyzed_results[C.F_y_by_x_in_fx] = f
      else:
        self.analyzed_results[C.F_y_by_x_in_fx] = None
    return self.analyzed_results[C.F_y_by_x_in_fx]

  @property
  def F_x_by_y_in_fy(self):
    if C.F_x_by_y_in_fy not in self.analyzed_results:
      if C.x_by_y_in_fy in self.options:
        wrapper = utils.std_func(utils.get_args(self.options[C.x_by_y_in_fy], gather_var=False),
                                 self.target_var_names[1:],
                                 self.target_par_names)
        f = wrapper(self.options[C.x_by_y_in_fy])
        f = partial(f, **(self.pars_update + self.fixed_vars))
        f = utils.f_without_jaxarray_return(f)
        self.analyzed_results[C.F_x_by_y_in_fy] = f
      else:
        self.analyzed_results[C.F_x_by_y_in_fy] = None
    return self.analyzed_results[C.F_x_by_y_in_fy]

  @property
  def F_y_by_x_in_fy(self):
    if C.F_y_by_x_in_fy not in self.analyzed_results:
      if C.y_by_x_in_fy in self.options:
        wrapper = utils.std_func(utils.get_args(self.options[C.y_by_x_in_fy], gather_var=False),
                                 self.target_var_names[:1] + self.target_var_names[2:],
                                 self.target_par_names)
        f = wrapper(self.options[C.y_by_x_in_fy])
        f = partial(f, **(self.pars_update + self.fixed_vars))
        f = utils.f_without_jaxarray_return(f)
        self.analyzed_results[C.F_y_by_x_in_fy] = f
      else:
        self.analyzed_results[C.F_y_by_x_in_fy] = None
    return self.analyzed_results[C.F_y_by_x_in_fy]

  @property
  def F_vmap_fy(self):
    if C.F_vmap_fy not in self.analyzed_results:
      self.analyzed_results[C.F_vmap_fy] = bm.jit(vmap(self.F_fy), device=self.jit_device)
    return self.analyzed_results[C.F_vmap_fy]

  @property
  def F_dfxdy(self):
    """The function to evaluate :math:`\frac{df_x (*\mathrm{vars}, *\mathrm{pars})}{dy}`."""
    if C.F_dfxdy not in self.analyzed_results:
      dfxdy = bm.vector_grad(self.F_fx, argnums=1)
      self.analyzed_results[C.F_dfxdy] = bm.jit(dfxdy, device=self.jit_device)
    return self.analyzed_results[C.F_dfxdy]

  @property
  def F_dfydx(self):
    """The function to evaluate :math:`\frac{df_y (*\mathrm{vars}, *\mathrm{pars})}{dx}`."""
    if C.F_dfydx not in self.analyzed_results:
      dfydx = bm.vector_grad(self.F_fy, argnums=0)
      self.analyzed_results[C.F_dfydx] = bm.jit(dfydx, device=self.jit_device)
    return self.analyzed_results[C.F_dfydx]

  @property
  def F_dfydy(self):
    """The function to evaluate :math:`\frac{df_y (*\mathrm{vars}, *\mathrm{pars})}{dy}`."""
    if C.F_dfydy not in self.analyzed_results:
      dfydy = bm.vector_grad(self.F_fy, argnums=1)
      self.analyzed_results[C.F_dfydy] = bm.jit(dfydy, device=self.jit_device)
    return self.analyzed_results[C.F_dfydy]

  @property
  def F_jacobian(self):
    """The function to evaluate :math:`J(*\mathrm{vars}, *\mathrm{pars})`."""
    if C.F_jacobian not in self.analyzed_results:
      @partial(bm.jacobian, argnums=(0, 1))
      def f_jacobian(*var_and_pars):
        return self.F_fx(*var_and_pars), self.F_fy(*var_and_pars)

      def call(*var_and_pars):
        var_and_pars = tuple((vp.value if isinstance(vp, bm.JaxArray) else vp) for vp in var_and_pars)
        return jnp.array(bm.jit(f_jacobian, device=self.jit_device)(*var_and_pars))

      self.analyzed_results[C.F_jacobian] = call
    return self.analyzed_results[C.F_jacobian]

  @property
  def F_fixed_point_opt(self):
    if C.F_fixed_point_opt not in self.analyzed_results:
      if self._can_convert_to_one_eq():
        if self.convert_type() == C.x_by_y:
          def f(start_and_end, *args):
            return utils.jax_brentq(self.F_y_convert[1])(start_and_end[0], start_and_end[1], args)
        else:
          def f(start_and_end, *args):
            return utils.jax_brentq(self.F_x_convert[1])(start_and_end[0], start_and_end[1], args)
        self.analyzed_results[C.F_fixed_point_opt] = f

      else:
        # If cannot convert to one variable equation
        def opt_fun(xy_init, *args):
          # "xy_init" is a vector with length 2,
          # "args: is a tuple of scalar
          return minimize(self.F_fixed_point_aux, xy_init, args=args, method='BFGS')

        self.analyzed_results[C.F_fixed_point_opt] = opt_fun
    return self.analyzed_results[C.F_fixed_point_opt]

  @property
  def F_fixed_point_aux(self):
    if C.F_fixed_point_aux not in self.analyzed_results:
      if self._can_convert_to_one_eq():
        if self.convert_type() == C.x_by_y:
          f = lambda y, *args: jnp.abs(self.F_y_convert[1](y, *args)).sum()
        else:
          f = lambda x, *args: jnp.abs(self.F_x_convert[1](x, *args)).sum()
        self.analyzed_results[C.F_fixed_point_aux] = f

      else:
        def aux_fun(xy, *args):
          # "xy" is a vector with length 2,
          # "args": is a tuple of scalar
          dx = self.F_fx(xy[0], xy[1], *args)
          dy = self.F_fy(xy[0], xy[1], *args)
          # return (jnp.abs(dx) + jnp.abs(dy)).sum()
          return (dx ** 2 + dy ** 2).sum()

        self.analyzed_results[C.F_fixed_point_aux] = aux_fun
    return self.analyzed_results[C.F_fixed_point_aux]

  def _can_convert_to_one_eq(self):
    if self.F_x_by_y_in_fx is not None:
      return True
    if self.F_x_by_y_in_fy is not None:
      return True
    if self.F_y_by_x_in_fx is not None:
      return True
    if self.F_y_by_x_in_fy is not None:
      return True
    return False

  def convert_type(self):
    if self.F_x_by_y_in_fx is not None:
      return C.x_by_y
    if self.F_x_by_y_in_fy is not None:
      return C.x_by_y
    if self.F_y_by_x_in_fx is not None:
      return C.y_by_x
    if self.F_y_by_x_in_fy is not None:
      return C.y_by_x
    raise errors.AnalyzerError

  @property
  def F_y_convert(self):
    if C.F_y_convert not in self.analyzed_results:
      if self.F_x_by_y_in_fy is not None:
        f = lambda y, *pars: self.F_fx(self.F_x_by_y_in_fy(y, *pars), y, *pars)
        res = (self.F_x_by_y_in_fy, f)
      elif self.F_x_by_y_in_fx is not None:
        f = lambda y, *pars: self.F_fy(self.F_x_by_y_in_fx(y, *pars), y, *pars)
        res = (self.F_x_by_y_in_fx, f)
      else:
        res = None
      self.analyzed_results[C.F_y_convert] = res
    return self.analyzed_results[C.F_y_convert]

  @property
  def F_x_convert(self):
    if C.F_x_convert not in self.analyzed_results:
      if self.F_y_by_x_in_fy is not None:
        f = lambda x, *pars: self.F_fx(x, self.F_y_by_x_in_fy(x, *pars), *pars)
        res = (self.F_y_by_x_in_fy, f)
      elif self.F_y_by_x_in_fx is not None:
        f = lambda x, *pars: self.F_fy(x, self.F_y_by_x_in_fx(x, *pars), *pars)
        res = (self.F_y_by_x_in_fx, f)
      else:
        res = None
      self.analyzed_results[C.F_x_convert] = res
    return self.analyzed_results[C.F_x_convert]

  def _fp_filter(self, x_values, y_values, par_values, aux_filter=0.):
    if aux_filter > 0.:
      losses = self.F_vmap_fp_aux(jnp.stack([x_values, y_values]).T, *par_values)
      ids = jnp.where(losses < aux_filter)[0]
      x_values = x_values[ids]
      y_values = y_values[ids]
      par_values = tuple(p[ids] for p in par_values)
    return x_values, y_values, par_values

  def _get_fx_nullcline_points(self, coords=None, tol=1e-7, num_segments=1, fp_aux_filter=0.):
    coords = (self.x_var + '-' + self.y_var) if coords is None else coords
    key = C.fx_nullcline_points + ',' + coords
    if key not in self.analyzed_results:
      all_losses = []
      all_x_values_in_fx = []
      all_y_values_in_fx = []
      all_p_values_in_fx = tuple([] for _ in range(len(self.target_par_names)))

      # points of variables and parameters
      xs = self.resolutions[self.x_var].value
      ys = self.resolutions[self.y_var].value
      par_seg = utils.Segment(targets=tuple(self.resolutions[p].value for p in self.target_par_names),
                              num_segments=num_segments)

      if self.F_x_by_y_in_fx is not None:
        utils.output("I am evaluating fx-nullcline by F_x_by_y_in_fx ...")
        vmap_f = bm.jit(vmap(self.F_x_by_y_in_fx), device=self.jit_device)
        for j, pars in enumerate(par_seg):
          if len(par_seg.arg_id_segments[0]) > 1: utils.output(f"{C.prefix}segment {j} ...")
          mesh_values = jnp.meshgrid(*((ys,) + pars))
          x_values_in_fx = vmap_f(*mesh_values)
          y_values_in_fx = mesh_values[0]
          p_values_in_fx = mesh_values[1:]
          losses = self.F_vmap_fx(x_values_in_fx, y_values_in_fx, *p_values_in_fx)
          all_losses.append(losses)
          all_x_values_in_fx.append(x_values_in_fx)
          all_y_values_in_fx.append(y_values_in_fx)
          for i, arg in enumerate(p_values_in_fx):
            all_p_values_in_fx[i].append(arg)

      elif self.F_y_by_x_in_fx is not None:
        utils.output("I am evaluating fx-nullcline by F_y_by_x_in_fx ...")
        vmap_f = bm.jit(vmap(self.F_y_by_x_in_fx), device=self.jit_device)
        for j, pars in enumerate(par_seg):
          if len(par_seg.arg_id_segments[0]) > 1: utils.output(f"{C.prefix}segment {j} ...")
          mesh_values = jnp.meshgrid(*((xs,) + pars))
          y_values_in_fx = vmap_f(*mesh_values)
          x_values_in_fx = mesh_values[0]
          p_values_in_fx = mesh_values[1:]
          losses = self.F_vmap_fx(x_values_in_fx, y_values_in_fx, *p_values_in_fx)
          all_losses.append(losses)
          all_x_values_in_fx.append(x_values_in_fx)
          all_y_values_in_fx.append(y_values_in_fx)
          for i, arg in enumerate(p_values_in_fx):
            all_p_values_in_fx[i].append(arg)

      else:
        utils.output("I am evaluating fx-nullcline by optimization ...")
        # auxiliary functions
        f2 = lambda y, x, *pars: self.F_fx(x, y, *pars)
        vmap_f2 = bm.jit(vmap(f2), device=self.jit_device)
        vmap_brentq_f2 = bm.jit(vmap(utils.jax_brentq(f2)), device=self.jit_device)
        vmap_brentq_f1 = bm.jit(vmap(utils.jax_brentq(self.F_fx)), device=self.jit_device)

        # num segments
        for _j, Ps in enumerate(par_seg):
          if len(par_seg.arg_id_segments[0]) > 1:
            utils.output(f"{C.prefix}segment {_j} ...")
          if coords == self.x_var + '-' + self.y_var:
            x0s, x1s, vps = utils.brentq_candidates(self.F_vmap_fx, *((xs, ys) + Ps))
            x_values_in_fx, out_args = utils.brentq_roots2(vmap_brentq_f1, x0s, x1s, *vps)
            y_values_in_fx = out_args[0]
            p_values_in_fx = out_args[1:]
            x_values_in_fx, y_values_in_fx, p_values_in_fx = \
              self._fp_filter(x_values_in_fx, y_values_in_fx, p_values_in_fx, fp_aux_filter)
          elif coords == self.y_var + '-' + self.x_var:
            x0s, x1s, vps = utils.brentq_candidates(vmap_f2, *((ys, xs) + Ps))
            y_values_in_fx, out_args = utils.brentq_roots2(vmap_brentq_f2, x0s, x1s, *vps)
            x_values_in_fx = out_args[0]
            p_values_in_fx = out_args[1:]
            x_values_in_fx, y_values_in_fx, p_values_in_fx = \
              self._fp_filter(x_values_in_fx, y_values_in_fx, p_values_in_fx, fp_aux_filter)
          else:
            raise ValueError
          losses = self.F_vmap_fx(x_values_in_fx, y_values_in_fx, *p_values_in_fx)
          all_losses.append(losses)
          all_x_values_in_fx.append(x_values_in_fx)
          all_y_values_in_fx.append(y_values_in_fx)
          for i, arg in enumerate(p_values_in_fx):
            all_p_values_in_fx[i].append(arg)

      all_losses = jnp.concatenate(all_losses)
      all_x_values_in_fx = jnp.concatenate(all_x_values_in_fx)
      all_y_values_in_fx = jnp.concatenate(all_y_values_in_fx)
      all_p_values_in_fx = tuple(jnp.concatenate(p) for p in all_p_values_in_fx)
      ids = jnp.where(all_losses < tol)[0]
      all_x_values_in_fx = all_x_values_in_fx[ids]
      all_y_values_in_fx = all_y_values_in_fx[ids]
      all_p_values_in_fx = tuple(a[ids] for a in all_p_values_in_fx)
      all_xy_values = jnp.stack([all_x_values_in_fx, all_y_values_in_fx]).T
      self.analyzed_results[key] = (all_xy_values,) + all_p_values_in_fx
    return self.analyzed_results[key]

  def _get_fy_nullcline_points(self, coords=None, tol=1e-7, num_segments=1, fp_aux_filter=0.):
    coords = (self.x_var + '-' + self.y_var) if coords is None else coords
    key = C.fy_nullcline_points + ',' + coords
    if key not in self.analyzed_results:
      all_losses = []
      all_x_values_in_fy = []
      all_y_values_in_fy = []
      all_p_values_in_fy = tuple([] for _ in range(len(self.target_par_names)))

      xs = self.resolutions[self.x_var].value
      ys = self.resolutions[self.y_var].value
      par_seg = utils.Segment(tuple(self.resolutions[p].value for p in self.target_par_names),
                              num_segments=num_segments)

      if self.F_x_by_y_in_fy is not None:
        utils.output("I am evaluating fy-nullcline by F_x_by_y_in_fy ...")
        vmap_f = bm.jit(vmap(self.F_x_by_y_in_fy), device=self.jit_device)
        for j, pars in enumerate(par_seg):
          if len(par_seg.arg_id_segments[0]) > 1: utils.output(f"{C.prefix}segment {j} ...")
          mesh_values = jnp.meshgrid(*((ys,) + pars))
          x_values_in_fy = vmap_f(*mesh_values)
          y_values_in_fy = mesh_values[0]
          p_values_in_fy = mesh_values[1:]
          losses = self.F_vmap_fy(x_values_in_fy, y_values_in_fy, *p_values_in_fy)
          all_losses.append(losses)
          all_x_values_in_fy.append(x_values_in_fy)
          all_y_values_in_fy.append(y_values_in_fy)
          for i, arg in enumerate(p_values_in_fy):
            all_p_values_in_fy[i].append(arg)

      elif self.F_y_by_x_in_fy is not None:
        utils.output("I am evaluating fy-nullcline by F_y_by_x_in_fy ...")
        vmap_f = bm.jit(vmap(self.F_y_by_x_in_fy), device=self.jit_device)
        for j, pars in enumerate(par_seg):
          if len(par_seg.arg_id_segments[0]) > 1: utils.output(f"{C.prefix}segment {j} ...")
          mesh_values = jnp.meshgrid(*((xs,) + pars))
          y_values_in_fy = vmap_f(*mesh_values)
          x_values_in_fy = mesh_values[0]
          p_values_in_fy = mesh_values[1:]
          losses = self.F_vmap_fy(x_values_in_fy, y_values_in_fy, *p_values_in_fy)
          all_losses.append(losses)
          all_x_values_in_fy.append(x_values_in_fy)
          all_y_values_in_fy.append(y_values_in_fy)
          for i, arg in enumerate(p_values_in_fy):
            all_p_values_in_fy[i].append(arg)

      else:
        utils.output("I am evaluating fy-nullcline by optimization ...")

        # auxiliary functions
        f2 = lambda y, x, *pars: self.F_fy(x, y, *pars)
        vmap_f2 = bm.jit(vmap(f2), device=self.jit_device)
        vmap_brentq_f2 = bm.jit(vmap(utils.jax_brentq(f2)), device=self.jit_device)
        vmap_brentq_f1 = bm.jit(vmap(utils.jax_brentq(self.F_fy)), device=self.jit_device)

        for j, Ps in enumerate(par_seg):
          if len(par_seg.arg_id_segments[0]) > 1: utils.output(f"{C.prefix}segment {j} ...")
          if coords == self.x_var + '-' + self.y_var:
            starts, ends, vps = utils.brentq_candidates(self.F_vmap_fy, *((xs, ys) + Ps))
            x_values_in_fy, out_args = utils.brentq_roots2(vmap_brentq_f1, starts, ends, *vps)
            y_values_in_fy = out_args[0]
            p_values_in_fy = out_args[1:]
            x_values_in_fy, y_values_in_fy, p_values_in_fy = \
              self._fp_filter(x_values_in_fy, y_values_in_fy, p_values_in_fy, fp_aux_filter)
          elif coords == self.y_var + '-' + self.x_var:
            starts, ends, vps = utils.brentq_candidates(vmap_f2, *((ys, xs) + Ps))
            y_values_in_fy, out_args = utils.brentq_roots2(vmap_brentq_f2, starts, ends, *vps)
            x_values_in_fy = out_args[0]
            p_values_in_fy = out_args[1:]
            x_values_in_fy, y_values_in_fy, p_values_in_fy = \
              self._fp_filter(x_values_in_fy, y_values_in_fy, p_values_in_fy, fp_aux_filter)
          else:
            raise ValueError
          losses = self.F_vmap_fy(x_values_in_fy, y_values_in_fy, *p_values_in_fy)
          all_losses.append(losses)
          all_x_values_in_fy.append(x_values_in_fy)
          all_y_values_in_fy.append(y_values_in_fy)
          for i, arg in enumerate(p_values_in_fy):
            all_p_values_in_fy[i].append(arg)
      all_losses = jnp.concatenate(all_losses)
      all_x_values_in_fy = jnp.concatenate(all_x_values_in_fy)
      all_y_values_in_fy = jnp.concatenate(all_y_values_in_fy)
      all_p_values_in_fy = tuple(jnp.concatenate(p) for p in all_p_values_in_fy)
      ids = jnp.where(all_losses < tol)[0]
      all_x_values_in_fy = all_x_values_in_fy[ids]
      all_y_values_in_fy = all_y_values_in_fy[ids]
      all_p_values_in_fy = tuple(a[ids] for a in all_p_values_in_fy)
      all_xy_values = jnp.stack([all_x_values_in_fy, all_y_values_in_fy]).T
      self.analyzed_results[key] = (all_xy_values,) + all_p_values_in_fy
    return self.analyzed_results[key]

  def _get_fp_candidates_by_aux_rank(self, num_segments=1, num_rank=100):
    utils.output(f"I am filtering out fixed point candidates with auxiliary function ...")
    all_xs = []
    all_ys = []
    all_ps = tuple([] for _ in range(len(self.target_par_names)))

    # points of variables and parameters
    xs = self.resolutions[self.x_var].value
    ys = self.resolutions[self.y_var].value
    P = tuple(self.resolutions[p].value for p in self.target_par_names)
    f_select = bm.jit(vmap(lambda vals, ids: vals[ids], in_axes=(1, 1)))

    # num seguments
    if isinstance(num_segments, int):
      num_segments = tuple([num_segments] * len(self.target_par_names))
    assert isinstance(num_segments, (tuple, list)) and len(num_segments) == len(self.target_par_names)
    arg_lens = tuple(len(p) for p in P)
    arg_pre_len = tuple(int(np.ceil(l / num_segments[i])) for i, l in enumerate(arg_lens))
    arg_id_segments = tuple(np.arange(0, l, arg_pre_len[i]) for i, l in enumerate(arg_lens))
    arg_id_segments = tuple(ids.flatten() for ids in np.meshgrid(*arg_id_segments))
    if len(arg_id_segments) == 0:
      arg_id_segments = ((0,),)
    for _j, ids in enumerate(zip(*arg_id_segments)):
      if len(arg_id_segments[0]) > 1:
        utils.output(f"{C.prefix}segment {_j} ...")

      ps = tuple(p[ids[i]: ids[i] + arg_pre_len[i]] for i, p in enumerate(P))
      # change the position of meshgrid values
      vps = tuple((v.value if isinstance(v, bm.JaxArray) else v) for v in ((xs, ys) + ps))
      mesh_values = jnp.meshgrid(*vps)
      mesh_values = tuple(jnp.moveaxis(m, 0, 1) for m in mesh_values)
      mesh_values = tuple(m.flatten() for m in mesh_values)
      # function outputs
      losses = self.F_vmap_fp_aux(jnp.stack([mesh_values[0], mesh_values[1]]).T, *mesh_values[2:])
      shape = (len(xs) * len(ys), -1)
      losses = losses.reshape(shape)
      argsorts = jnp.argsort(losses, axis=0)[:num_rank]
      all_xs.append(f_select(mesh_values[0].reshape(shape), argsorts).flatten())
      all_ys.append(f_select(mesh_values[1].reshape(shape), argsorts).flatten())
      for i, p in enumerate(ps):
        all_ps[i].append(f_select(mesh_values[i + 2].reshape(shape), argsorts).flatten())
    all_xys = jnp.vstack([jnp.concatenate(all_xs), jnp.concatenate(all_ys)]).T
    all_ps = tuple(jnp.concatenate(p) for p in all_ps)
    return (all_xys, all_ps)

  def _get_fixed_points(self, candidates, *args, tol_aux=1e-7,
                        tol_unique=1e-2, tol_opt_candidate=None,
                        num_segment=1):
    """Get the fixed points according to the initial ``candidates`` and the parameter setting ``args``.

    "candidates" and "args" can be obtained through:

    >>> all_candidates = []
    >>> all_par1 = []
    >>> all_par2 = []
    >>> for p1 in par1_list:
    >>>   for p2 in par2_list:
    >>>     nullcline_points = _get_nullcline_points(p1, p2)
    >>>     all_candidates.append(nullcline_points)
    >>>     all_par1.append(jnp.ones_like(nullcline_points) * p1)
    >>>     all_par2.append(jnp.ones_like(nullcline_points) * p2)

    Parameters
    ----------
    candidates: np.ndarray, jnp.ndarray
      The candidate points (batched) to optimize, like the nullcline points.
    args : tuple
      The parameters (batched).
    tol_aux : float
    tol_unique : float
    tol_opt_candidate : float, optional

    Returns
    -------
    res : tuple
      The fixed point results.
    """

    if self._can_convert_to_one_eq():
      utils.output("I am trying to find fixed points by brentq optimization ...")

      # candidates: xs, a vector with the length of self.resolutions[self.x_var]
      # args: parameters, a list/tuple of vectors
      candidates = candidates.value if isinstance(candidates, bm.JaxArray) else candidates
      selected_ids = np.arange(len(candidates))
      args = tuple(a.value if isinstance(candidates, bm.JaxArray) else a for a in args)
      for a in args: assert len(a) == len(candidates)

      if self.convert_type() == C.x_by_y:
        num_seg = len(self.resolutions[self.y_var])
        f_vmap = bm.jit(vmap(self.F_y_convert[1]))
      else:
        num_seg = len(self.resolutions[self.x_var])
        f_vmap = bm.jit(vmap(self.F_x_convert[1]))
      # get the signs
      signs = jnp.sign(f_vmap(candidates, *args))
      signs = signs.reshape((num_seg, -1))
      par_len = signs.shape[1]
      signs1 = signs.at[-1].set(1)
      signs2 = jnp.vstack((signs[1:], signs[:1])).at[-1].set(1)
      ids = jnp.where((signs1 * signs2).flatten() <= 0)[0]
      if len(ids) <= 0:
        return [], [], []

      # selected the proper candidates to optimize fixed points
      selected_ids = selected_ids[np.asarray(ids)]
      starts = candidates[ids]
      ends = candidates[ids + par_len]
      X = jnp.stack((starts, ends)).T
      args = tuple(a[ids] for a in args)

      # optimize the fixed points
      res = self.F_vmap_fp_opt(X, *args)
      losses = self.F_vmap_fp_aux(res['root'], *args)
      valid_or_not = jnp.logical_and(res['status'] == utils.ECONVERGED, losses <= tol_aux)
      ids = np.asarray(jnp.where(valid_or_not)[0])
      fps = np.asarray(res['root'])[ids]
      args = tuple(a[ids] for a in args)
      selected_ids = selected_ids[np.asarray(ids)]

      # get another value
      if self.convert_type() == C.x_by_y:
        y_values = fps
        x_values = bm.jit(vmap(self.F_y_convert[0]))(y_values, *args)
      else:
        x_values = fps
        y_values = bm.jit(vmap(self.F_x_convert[0]))(x_values, *args)
      fps = jnp.stack([x_values, y_values]).T
      return fps, selected_ids, args

    else:
      utils.output("I am trying to find fixed points by optimization ...")
      utils.output(f"{C.prefix}There are {len(candidates)} candidates")

      candidates = jnp.asarray(candidates)
      args = tuple(jnp.asarray(a) for a in args)

      all_ids = []
      all_fps = []
      all_args = tuple([] for _ in range(len(args)))
      seg_len = int(np.ceil(len(candidates) / num_segment))
      segment_ids = np.arange(0, len(candidates), seg_len)
      selected_ids = jnp.arange(len(candidates))

      for _j, i in enumerate(segment_ids):
        if len(segment_ids) > 1:
          utils.output(f"{C.prefix}segment {_j} ...")
        seg_fps = candidates[i: i + seg_len]
        seg_args = tuple(a[i: i + seg_len] for a in args)
        seg_ids = selected_ids[i: i + seg_len]

        if tol_opt_candidate is not None:
          # screen by the function loss
          losses = self.F_vmap_fp_aux(seg_fps, *seg_args)
          ids = jnp.where(losses < tol_opt_candidate)[0]
          seg_fps = seg_fps[ids]
          seg_args = tuple(a[ids] for a in seg_args)
          seg_ids = seg_ids[ids]
        if len(seg_fps):
          # optimization
          seg_fps = self.F_vmap_fp_opt(seg_fps, *seg_args)
          # loss
          losses = self.F_vmap_fp_aux(seg_fps.x, *seg_args)
          # valid indices
          ids = jnp.where(losses <= tol_aux)[0]
          seg_ids = seg_ids[ids]
          all_fps.append(seg_fps.x[ids])
          all_ids.append(seg_ids)
          for i in range(len(all_args)):
            all_args[i].append(seg_args[i][ids])
      all_fps = jnp.concatenate(all_fps)
      all_ids = jnp.concatenate(all_ids)
      all_args = tuple(jnp.concatenate(args) for args in all_args)
      return all_fps, all_ids, all_args


class Num3DAnalyzer(Num2DAnalyzer):
  def __init__(self, *args, **kwargs):
    super(Num3DAnalyzer, self).__init__(*args, **kwargs)
    if len(self.target_vars) < 3:
      raise errors.AnalyzerError(f'{Num1DAnalyzer.__name__} only supports dynamical system '
                                 f'with >= 3 variables. But we got {len(self.target_vars)} '
                                 f'variables in {self.model}.')
    self.z_var = self.target_var_names[2]

  @property
  def F_fz(self):
    """The function to evaluate :math:`f_y(*\mathrm{vars}, *\mathrm{pars})`."""
    if C.F_fz not in self.analyzed_results:
      variables, arguments = utils.get_args(self.model.F[self.z_var])
      wrapper = utils.std_derivative(arguments, self.target_var_names, self.target_par_names)
      f = wrapper(self.model.F[self.z_var])
      f = partial(f, **(self.pars_update + self.fixed_vars))
      self.analyzed_results[C.F_fz] = bm.jit(f, device=self.jit_device)
    return self.analyzed_results[C.F_fz]

  def fz_signs(self, pars=(), cache=False):
    xyz = tuple(self.resolutions.values())
    return utils.get_sign2(self.F_fz, *xyz, args=pars)



# -*- coding: utf-8 -*-

import jax.numpy as jnp

from brainpy import math as bm, dyn
from brainpy.integrators import odeint, ddeint, JointEq, IntegratorRunner

__all__ = [
  'henon_map_series',
  'logistic_map_series',
  'modified_lu_chen_series',
  'mackey_glass_series',

  'rabinovich_fabrikant_series',
  'chen_chaotic_series',
  'lu_chen_chaotic_series',
  'chua_chaotic_series',
  'modified_chua_series',
  'lorenz_series',
  'modified_Lorenz_series',
  'double_scroll_series',

  'PWL_duffing_series',
]


def henon_map_series(num_step, a=1.4, b=0.3, inits=None, numpy_mon=False):
  r"""The Hénon map time series.

  The Hénon map is a discrete-time dynamical system. It is one of the
  most studied examples of dynamical systems that exhibit chaotic behavior.

  .. math::

    \begin{split}\begin{cases}x_{n+1} = 1-a x_n^2 + y_n\\y_{n+1} = b x_n.\end{cases}\end{split}

  The map depends on two parameters, a and b, which for the classical
  Hénon map have values of a = 1.4 and b = 0.3. For the classical values
  the Hénon map is chaotic.

  References
  ----------

  .. [1] https://brainpy-examples.readthedocs.io/en/latest/classical_dynamical_systems/henon_map.html
  .. [1] https://en.wikipedia.org/wiki/H%C3%A9non_map
  """
  if inits is None:
    inits = {'x': bm.zeros(1), 'y': bm.zeros(1)}
  elif isinstance(inits, dict):
    assert 'x' in inits
    assert 'y' in inits
    inits = {'x': bm.asarray(inits['x']), 'y': bm.asarray(inits['y'])}
    assert inits['x'].shape == inits['y'].shape
  else:
    raise ValueError(f'Please provide dict, and do not support {type(inits)}: {inits}')
  map = _HenonMap(inits['x'].size, a=a, b=b)
  runner = dyn.DSRunner(map, monitors=['x', 'y'], dt=1, progress_bar=False,
                        numpy_mon_after_run=numpy_mon)
  runner.run(num_step)
  return {'ts': runner.mon.ts,
          'x': runner.mon.x,
          'y': runner.mon.y}


def logistic_map_series(num_step, mu=3., inits=None, numpy_mon=False):
  r"""The logistic map time series.

  The logistic map is defined by the following equation:

  .. math::

     x_{n+1}=\lambda x_{n}\left(1-x_{n}\right) \quad \text { with } \quad n=0,1,2,3 \ldots

  References
  ----------
  .. [3] https://brainpy-examples.readthedocs.io/en/latest/classical_dynamical_systems/logistic_map.html
  .. [4] https://en.wikipedia.org/wiki/Logistic_map

  """
  if inits is None:
    inits = bm.ones(1) * 0.2
  else:
    inits = bm.asarray(inits)
  runner = dyn.DSRunner(_LogisticMap(inits.size, mu=mu),
                        monitors=['x'], dt=1, progress_bar=False,
                        numpy_mon_after_run=numpy_mon)
  runner.run(num_step)
  return {'ts': runner.mon.ts, 'x': runner.mon.x}


def modified_lu_chen_series(duration, dt=0.001, a=36, c=20, b=3, d1=1, d2=0., tau=.2,
                            method='rk4', inits=None, numpy_mon=False):
  """Modified Lu Chen attractor.

  References
  ----------
  .. [4] https://brainpy-examples.readthedocs.io/en/latest/classical_dynamical_systems/Multiscroll_attractor.html#Modified-Lu-Chen-attractor
  """
  if inits is None:
    inits = {'x': bm.ones(1), 'y': bm.ones(1), 'z': bm.ones(1) * 14}
  elif isinstance(inits, dict):
    assert 'x' in inits
    assert 'y' in inits
    assert 'z' in inits
    inits = {'x': bm.asarray(inits['x']),
             'y': bm.asarray(inits['y']),
             'z': bm.asarray(inits['z'])}
    assert inits['x'].shape == inits['y'].shape == inits['z'].shape
  else:
    raise ValueError
  eq = _ModifiedLuChenSystem(num=inits['x'].size, a=a, b=b, c=c, d1=d1, d2=d2, tau=tau, dt=dt, method=method)
  eq.x[:] = inits['x']
  eq.y[:] = inits['y']
  eq.z[:] = inits['z']
  runner = dyn.DSRunner(eq,
                        monitors=['x', 'y', 'z'],
                        dt=dt, progress_bar=False,
                        numpy_mon_after_run=numpy_mon)
  runner.run(duration)
  return {'ts': runner.mon.ts,
          'x': runner.mon['x'],
          'y': runner.mon['y'],
          'z': runner.mon['z']}


def mackey_glass_series(duration, dt=0.1, beta=2., gamma=1., tau=2., n=9.65,
                        inits=None, method='rk4', seed=None,
                        progress_bar=False, numpy_mon=False):
  """The Mackey-Glass time series.

  Its dynamics is governed by

  .. math::

     \frac{dP(t)}{dt} = \frac{\beta P(t - \tau)}{1 + P(t - \tau)^n} - \gamma P(t)

  where $\beta = 0.2$, $\gamma = 0.1$, $n = 10$, and the time delay $\tau = 17$. $\tau$
  controls the chaotic behaviour of the equations (the higher it is, the more chaotic
  the timeserie becomes.)

  Parameters
  ----------
  duration: int
  dt: float, int, optional
  beta: float, JaxArray
  gamma: float, JaxArray
  tau: float, JaxArray
  n: float, JaxArray
  inits: optional, float, JaxArray
  method: str
  seed: optional, int
  progress_bar: bool

  Returns
  -------
  result: dict
    The time series data which contain

  References
  ----------

  .. [5] https://brainpy-examples.readthedocs.io/en/latest/classical_dynamical_systems/mackey_glass_eq.html
  """
  if inits is None:
    inits = bm.ones(1) * 1.2
  elif isinstance(inits, (float, int)):
    inits = bm.asarray([inits], dtype=bm.float_)
  else:
    assert isinstance(inits, (bm.ndarray, jnp.ndarray))

  rng = bm.random.RandomState(seed)
  xdelay = bm.TimeDelay(inits, tau, dt=dt, interp_method='round')
  xdelay.data.value = inits + 0.2 * (rng.random((xdelay.num_delay_step,) + inits.shape) - 0.5)

  @ddeint(method=method,
          state_delays={'x': xdelay})
  def mg_eq(x, t):
    xtau = xdelay(t - tau)
    return beta * xtau / (1 + xtau ** n) - gamma * x

  runner = IntegratorRunner(mg_eq,
                            inits={'x': inits},
                            monitors=['x'],
                            fun_monitors={'x(tau)': lambda t, _: xdelay(t - tau)},
                            progress_bar=progress_bar, dt=dt,
                            numpy_mon_after_run=numpy_mon)
  runner.run(duration)
  return {'ts': runner.mon.ts,
          'x': runner.mon['x'],
          'x_tau': runner.mon['x(tau)']}


def lorenz_series(duration, dt=0.001, sigma=10, beta=8 / 3, rho=28, method='rk4', inits=None,
                  numpy_mon=False):
  """The Lorenz system.

  The Lorenz system is a system of ordinary differential equations first
  studied by mathematician and meteorologist Edward Lorenz.


  Returns
  -------
  data: dict
    A dict data with the keys of ``ts``, ``x``, ``y``, and ``z``,
    where ``ts`` is the history time value, ``x, y, z`` are history
    values of the variable in the Lorenz system.

  References
  ----------
  .. [6] https://brainpy-examples.readthedocs.io/en/latest/classical_dynamical_systems/lorenz_system.html

  """
  dx = lambda x, t, y: sigma * (y - x)
  dy = lambda y, t, x, z: x * (rho - z) - y
  dz = lambda z, t, x, y: x * y - beta * z
  integral = odeint(JointEq([dx, dy, dz]), method=method)

  return _three_variable_model(integral,
                               default_inits={'x': 8, 'y': 1, 'z': 1},
                               duration=duration, dt=dt, inits=inits,
                               numpy_mon=numpy_mon)


def rabinovich_fabrikant_series(duration, dt=0.001, alpha=1.1, gamma=0.803,
                                method='rk4', inits=None, numpy_mon=False):
  """Rabinovich-Fabrikant equations.

  The Rabinovich–Fabrikant equations are a set of three coupled ordinary
  differential equations exhibiting chaotic behaviour for certain values
  of the parameters. They are named after Mikhail Rabinovich and Anatoly
  Fabrikant, who described them in 1979.

  References
  ----------
  .. [7] https://brainpy-examples.readthedocs.io/en/latest/classical_dynamical_systems/Rabinovich_Fabrikant_eq.html

  """
  @odeint(method=method)
  def rf_eqs(x, y, z, t):
    dx = y * (z - 1 + x * x) + gamma * x
    dy = x * (3 * z + 1 - x * x) + gamma * y
    dz = -2 * z * (alpha + x * y)
    return dx, dy, dz

  return _three_variable_model(rf_eqs,
                               default_inits={'x': -1, 'y': 0, 'z': 0.5},
                               duration=duration, dt=dt, inits=inits,
                               numpy_mon=numpy_mon)


def chen_chaotic_series(duration, dt=0.001, a=40, b=3, c=28,
                        method='euler', inits=None, numpy_mon=False):
  """Chen attractor.

  References
  ----------
  .. [7] https://brainpy-examples.readthedocs.io/en/latest/classical_dynamical_systems/Multiscroll_attractor.html#Chen-attractor
  """

  @odeint(method=method)
  def chen_system(x, y, z, t):
    dx = a * (y - x)
    dy = (c - a) * x - x * z + c * y
    dz = x * y - b * z
    return dx, dy, dz

  return _three_variable_model(chen_system,
                               default_inits=dict(x=-0.1, y=0.5, z=-0.6),
                               duration=duration, dt=dt, inits=inits,
                               numpy_mon=numpy_mon)


def lu_chen_chaotic_series(duration, dt=0.001, a=36, c=20, b=3, u=-15.15,
                           method='rk4', inits=None, numpy_mon=False):
  """Lu Chen attractor.

  References
  ----------
  .. [8] https://brainpy-examples.readthedocs.io/en/latest/classical_dynamical_systems/Multiscroll_attractor.html#Lu-Chen-attractor
  """
  @odeint(method=method)
  def lu_chen_system(x, y, z, t):
    dx = a * (y - x)
    dy = x - x * z + c * y + u
    dz = x * y - b * z
    return dx, dy, dz

  return _three_variable_model(lu_chen_system,
                               default_inits=dict(x=0.1, y=0.3, z=-0.6),
                               duration=duration, dt=dt, inits=inits,
                               numpy_mon=numpy_mon)


def chua_chaotic_series(duration, dt=0.001, alpha=10, beta=14.514, gamma=0, a=-1.197, b=-0.646, method='rk4',
                        inits=None, numpy_mon=False):
  """Chua’s system.

  References
  ----------
  .. [9] https://brainpy-examples.readthedocs.io/en/latest/classical_dynamical_systems/Multiscroll_attractor.html#Chua%E2%80%99s-system
  """
  @odeint(method=method)
  def chua_equation(x, y, z, t):
    fx = b * x + 0.5 * (a - b) * (bm.abs(x + 1) - bm.abs(x - 1))
    dx = alpha * (y - x) - alpha * fx
    dy = x - y + z
    dz = -beta * y - gamma * z
    return dx, dy, dz

  return _three_variable_model(chua_equation,
                               default_inits=dict(x=0.001, y=0, z=0.),
                               duration=duration, dt=dt, inits=inits,
                               numpy_mon=numpy_mon)


def modified_chua_series(duration, dt=0.001, alpha=10.82, beta=14.286, a=1.3, b=.11, d=0,
                         method='rk4', inits=None, numpy_mon=False):
  """Modified Chua chaotic attractor.

  References
  ----------
  .. [10] https://brainpy-examples.readthedocs.io/en/latest/classical_dynamical_systems/Multiscroll_attractor.html#Modified-Chua-chaotic-attractor

  """

  @odeint(method=method)
  def modified_chua_system(x, y, z, t):
    dx = alpha * (y + b * bm.sin(bm.pi * x / 2 / a + d))
    dy = x - y + z
    dz = -beta * y
    return dx, dy, dz

  return _three_variable_model(modified_chua_system,
                               default_inits=dict(x=1, y=1, z=0.),
                               duration=duration, dt=dt, inits=inits,
                               numpy_mon=numpy_mon)


def modified_Lorenz_series(duration, dt=0.001, a=10, b=8 / 3, c=137 / 5,
                           method='rk4', inits=None, numpy_mon=False):
  """Modified Lorenz chaotic system.

  References
  ----------
  .. [11] https://brainpy-examples.readthedocs.io/en/latest/classical_dynamical_systems/Multiscroll_attractor.html#Modified-Lorenz-chaotic-system
  """

  @odeint(method=method)
  def modified_Lorenz(x, y, z, t):
    temp = 3 * bm.sqrt(x * x + y * y)
    dx = (-(a + 1) * x + a - c + z * y) / 3 + ((1 - a) * (x * x - y * y) + (2 * (a + c - z)) * x * y) / temp
    dy = ((c - a - z) * x - (a + 1) * y) / 3 + (2 * (a - 1) * x * y + (a + c - z) * (x * x - y * y)) / temp
    dz = (3 * x * x * y - y * y * y) / 2 - b * z
    return dx, dy, dz

  return _three_variable_model(modified_Lorenz,
                               default_inits=dict(x=-8, y=4, z=10),
                               duration=duration, dt=dt, inits=inits,
                               numpy_mon=numpy_mon)




def PWL_duffing_series(duration, dt=0.001, e=0.25, m0=-0.0845, m1=0.66, omega=1, i=-14,
                       method='rk4', inits=None, numpy_mon=False):
  """PWL Duffing chaotic attractor.

  References
  ----------
  .. [12] https://brainpy-examples.readthedocs.io/en/latest/classical_dynamical_systems/Multiscroll_attractor.html#PWL-Duffing-chaotic-attractor
  """

  gamma = 0.14 + i / 20

  @odeint(method=method)
  def PWL_duffing_eq(x, y, t):
    dx = y
    dy = -m1 * x - (0.5 * (m0 - m1)) * (abs(x + 1) - abs(x - 1)) - e * y + gamma * bm.cos(omega * t)
    return dx, dy

  return _two_variable_model(PWL_duffing_eq,
                             default_inits=dict(x=0, y=0.),
                             duration=duration, dt=dt, inits=inits,
                             numpy_mon=numpy_mon)


def double_scroll_series(duration, dt=0.01,
                         R1=1.2, R2=3.44, R4=0.193, beta=11.6, Ir=2 * 2.25e-5,
                         method='rk4', inits=None, numpy_mon=False):
  r"""Double-scroll electronic circuit attractor.

  Its behavior is governed by

  .. math::

     {\dot{V}}_{1} ={V}_{1}/{R}_{1}-\varDelta V/{R}_{2}\,-\,2{I}_{r}\,\sinh (\beta \varDelta V),\\
     \dot{{V}_{2}} =\varDelta V/{R}_{2}+2{I}_{r}\,\sinh (\beta \varDelta V)-I,\\
     \dot{I} ={V}_{2}-{R}_{4}I

  in dimensionless form.

  References
  ----------
  .. [1] Chang, A., Bienfang, J. C., Hall, G. M., Gardner, J. R. &
         Gauthier, D. J. Stabilizing unstable steady states using
         extended time-delay autosynchronization. Chaos 8, 782–790 (1998).
  """

  @odeint(method=method)
  def double_scroll(x, y, z, t):
    delta = x - y
    dV1 = x / R1 - delta / R2 - 2 * Ir * bm.sinh(beta * delta)
    dV2 = delta / R2 + 2 * Ir * bm.sinh(beta * delta) - z
    dI = y - R4 * z
    return dV1, dV2, dI

  return _three_variable_model(double_scroll,
                               default_inits=dict(x=0.37926545, y=0.058339, z=-0.08167691),
                               duration=duration, dt=dt, inits=inits,
                               numpy_mon=numpy_mon)


def _two_variable_model(integrator, duration, default_inits, inits=None,
                        args=None, dyn_args=None, dt=0.001, numpy_mon=False):
  if inits is None:
    inits = default_inits
  elif isinstance(inits, dict):
    assert 'x' in inits
    assert 'y' in inits
    inits = {'x': bm.asarray(inits['x']).flatten(),
             'y': bm.asarray(inits['y']).flatten()}
    assert inits['x'].shape == inits['y'].shape
  else:
    raise ValueError

  runner = IntegratorRunner(integrator, monitors=['x', 'y'], inits=inits,
                            args=args, dyn_args=dyn_args, dt=dt, progress_bar=False,
                            numpy_mon_after_run=numpy_mon)
  runner.run(duration)
  return {'ts': runner.mon.ts,
          'x': runner.mon.x,
          'y': runner.mon.y}


def _three_variable_model(integrator, duration, default_inits, inits=None, args=None,
                          dyn_args=None, dt=0.001, numpy_mon=False):
  if inits is None:
    inits = default_inits  # {'x': -1, 'y': 0, 'z': 0.5}
  elif isinstance(inits, dict):
    assert 'x' in inits
    assert 'y' in inits
    assert 'z' in inits
    inits = {'x': bm.asarray(inits['x']).flatten(),
             'y': bm.asarray(inits['y']).flatten(),
             'z': bm.asarray(inits['z']).flatten()}
    assert inits['x'].shape == inits['y'].shape == inits['z'].shape
  else:
    raise ValueError

  runner = IntegratorRunner(integrator, monitors=['x', 'y', 'z'], inits=inits,
                            args=args, dyn_args=dyn_args, dt=dt, progress_bar=False,
                            numpy_mon_after_run=numpy_mon)
  runner.run(duration)
  return {'ts': runner.mon.ts,
          'x': runner.mon.x,
          'y': runner.mon.y,
          'z': runner.mon.z}


class _HenonMap(dyn.DynamicalSystem):
  """Hénon map."""

  def __init__(self, num, a=1.4, b=0.3):
    super(_HenonMap, self).__init__()

    # parameters
    self.a = a
    self.b = b
    self.num = num

    # variables
    self.x = bm.Variable(bm.zeros(num))
    self.y = bm.Variable(bm.zeros(num))

  def update(self, _t, _dt):
    x_new = 1 - self.a * self.x * self.x + self.y
    self.y.value = self.b * self.x
    self.x.value = x_new


class _LogisticMap(dyn.DynamicalSystem):
  def __init__(self, num, mu=3.):
    super(_LogisticMap, self).__init__()

    self.mu = mu
    self.x = bm.Variable(bm.ones(num) * 0.2)

  def update(self, _t, _dt):
    self.x.value = self.mu * self.x * (1 - self.x)


class _ModifiedLuChenSystem(dyn.DynamicalSystem):
  def __init__(self, num, a=35, b=3, c=28, d0=1, d1=1, d2=0., tau=.2, dt=0.1, method='rk4'):
    super(_ModifiedLuChenSystem, self).__init__()

    # parameters
    self.a = a
    self.b = b
    self.c = c
    self.d0 = d0
    self.d1 = d1
    self.d2 = d2
    self.tau = tau

    # variables
    self.z_delay = dyn.ConstantDelay(num, delay=tau, dt=dt)
    self.z_delay.data[:] = 14
    self.z = bm.Variable(self.z_delay.latest)
    self.x = bm.Variable(bm.ones(num))
    self.y = bm.Variable(bm.ones(num))

    # functions
    def derivative(x, y, z, t):
      dx = self.a * (y - x)
      f = self.d0 * z + self.d1 * self.z_delay.pull() - self.d2 * bm.sin(self.z_delay.pull())
      dy = (self.c - self.a) * x - x * f + self.c * y
      dz = x * y - self.b * z
      return dx, dy, dz

    self.integral = odeint(derivative, method=method)

  def update(self, _t, _dt):
    self.x.value, self.y.value, self.z.value = self.integral(
      self.x, self.y, self.z, _t, _dt)
    self.z_delay.push(self.z)
    self.z_delay.update(_t, _dt)



# -*- coding: utf-8 -*-

import inspect

from brainpy import errors
from brainpy.base import Collector

__all__ = [
  'JointEq',
]


def _get_args(f):
  """Get the function arguments"""
  args = []
  kwargs = {}
  for name, par in inspect.signature(f).parameters.items():
    if par.kind is inspect.Parameter.POSITIONAL_OR_KEYWORD:
      if par.default is inspect._empty:
        args.append(par.name)
      else:
        kwargs[par.name] = par.default
    elif par.kind is inspect.Parameter.VAR_POSITIONAL:
      raise errors.DiffEqError(f'{JointEq.__name__} does not support VAR_POSITIONAL parameters '
                               f'*{par.name} (error in {f}).')
    elif par.kind is inspect.Parameter.KEYWORD_ONLY:
      raise errors.DiffEqError(f'{JointEq.__name__} does not support KEYWORD_ONLY parameters, '
                               f'e.g., *  (error in {f}).')
    elif par.kind is inspect.Parameter.POSITIONAL_ONLY:
      raise errors.DiffEqError(f'{JointEq.__name__} does not support POSITIONAL_ONLY parameters, '
                               'e.g., /  (error in {f}).')
    elif par.kind is inspect.Parameter.VAR_KEYWORD:
      raise errors.DiffEqError(f'{JointEq.__name__} does not support VAR_KEYWORD '
                               f'arguments **{par.name} (error in {f}).')
    else:
      raise errors.DiffEqError(f'Unknown argument type: {par.kind}')

  # variables
  vars = []
  for a in args:
    if a == 't':
      break
    vars.append(a)
  else:
    raise ValueError('Do not find time variable "t".')

  return vars, args, kwargs


def _std_func(f, all_vars: list):
  f_vars, f_args, f_kwargs = _get_args(f)

  def call(t, *vars, **args_and_kwargs):
    params = dict(t=t)
    for var in f_vars:
      params[var] = vars[all_vars.index(var)]
    for par in f_args[len(f_vars) + 1:]:
      if par in args_and_kwargs:
        params[par] = args_and_kwargs[par]
      else:
        if par not in all_vars:
          raise errors.DiffEqError(f'Missing {par} during the functional call of {f}.')
        params[par] = vars[all_vars.index(par)]
    for par, value in f_kwargs.items():
      if par in args_and_kwargs:
        params[par] = args_and_kwargs[par]
    return f(**params)

  return call


class JointEq(object):
  """Make a joint equation from multiple derivation functions.

  For example, we have an Izhikevich neuron model,

  >>> a, b = 0.02, 0.20
  >>> dV = lambda V, t, u, Iext: 0.04 * V * V + 5 * V + 140 - u + Iext
  >>> du = lambda u, t, V: a * (b * V - u)

  If we make numerical solver for each derivative function, they will be solved independently.

  >>> import brainpy as bp
  >>> bp.odeint(dV, method='rk2', show_code=True)
  def brainpy_itg_of_ode0(V, t, u, Iext, dt=0.1):
    dV_k1 = f(V, t, u, Iext)
    k2_V_arg = V + dt * dV_k1 * 0.6666666666666666
    k2_t_arg = t + dt * 0.6666666666666666
    dV_k2 = f(k2_V_arg, k2_t_arg, u, Iext)
    V_new = V + dV_k1 * dt * 0.25 + dV_k2 * dt * 0.75
    return V_new

  As you see in the output code, "dV_k2" is evaluated by :math:`f(V_{k2}, u)`.
  If you want to solve the above coupled equation jointly, i.e., evalute "dV_k2"
  with :math:`f(V_{k2}, u_{k2})`, you can use :py:class:`brainpy.JointEq`
  to emerge the above two derivative equations into a joint equation, so that
  they will be numerically solved together. Let's see the difference:

  >>> eq = bp.JointEq(eqs=(dV, du))
  >>> bp.odeint(eq, method='rk2', show_code=True)
  def brainpy_itg_of_ode0_joint_eq(V, u, t, Iext, dt=0.1):
    dV_k1, du_k1 = f(V, u, t, Iext)
    k2_V_arg = V + dt * dV_k1 * 0.6666666666666666
    k2_u_arg = u + dt * du_k1 * 0.6666666666666666
    k2_t_arg = t + dt * 0.6666666666666666
    dV_k2, du_k2 = f(k2_V_arg, k2_u_arg, k2_t_arg, Iext)
    V_new = V + dV_k1 * dt * 0.25 + dV_k2 * dt * 0.75
    u_new = u + du_k1 * dt * 0.25 + du_k2 * dt * 0.75
    return V_new, u_new

  :py:class:`brainpy.JointEq` supports make nested ``JointEq``, which means
  the instance of ``JointEq`` can be an element to compose a new ``JointEq``.

  >>> dw = lambda w, t, V: a * (b * V - w)
  >>> eq2 = bp.JointEq(eqs=(eq, dw))


  Parameters
  ----------
  eqs : sequence of function, sequence of callable
    The elements of derivative function to compose.
  """

  def __init__(self, eqs):
    # equations
    if not isinstance(eqs, (tuple, list)):
      raise errors.DiffEqError(f'"eqs" only supports list/tuple of '
                               f'derivative functions, but got {eqs}.')
    for eq in eqs:
      if not callable(eq):
        raise errors.DiffEqError(f'"eqs" only supports list/tuple of '
                                 f'derivative functions, but got {eq}.')

    # variables in equations
    self.vars_in_eqs = []
    vars_in_eqs = []
    for eq in eqs:
      vars, _, _ = _get_args(eq)
      for var in vars:
        if var in vars_in_eqs:
          raise errors.DiffEqError(f'Variable "{var}" has been used, however we got a same '
                                   f'variable name in {eq}. Please change another name.')
      vars_in_eqs.extend(vars)
      self.vars_in_eqs.append(vars)

    # arguments in equations
    self.args_in_eqs = []
    all_arg_pars = []
    all_kwarg_pars = dict()
    for eq in eqs:
      vars, args, kwargs = _get_args(eq)
      self.args_in_eqs.append(args + list(kwargs.keys()))
      for par in args[len(vars) + 1:]:
        if (par not in vars_in_eqs) and (par not in all_arg_pars) and (par not in all_kwarg_pars):
          all_arg_pars.append(par)
      for key, value in kwargs.items():
        if key in all_kwarg_pars and value != all_kwarg_pars[key]:
          raise errors.DiffEqError(f'We got two different default value of "{key}": '
                                   f'{all_kwarg_pars[key]} != {value}')
        elif (key not in vars_in_eqs) and (key not in all_arg_pars):
          all_kwarg_pars[key] = value
        else:
          raise errors.DiffEqError

    # # variable names provided
    # if not isinstance(variables, (tuple, list)):
    #   raise errors.DiffEqError(f'"variables" must be a list/tuple of str, but we got {variables}')
    # for v in variables:
    #   if not isinstance(v, str):
    #     raise errors.DiffEqError(f'"variables" must be a list/tuple of str, but we got {v} in "variables"')
    # if len(vars_in_eqs) != len(variables):
    #   raise errors.DiffEqError(f'We detect {len(vars_in_eqs)} variables "{vars_in_eqs}" '
    #                            f'in the provided equations. However, the used provided '
    #                            f'"variables" have {len(variables)} variables '
    #                            f'"{variables}".')
    # if len(set(vars_in_eqs) - set(variables)) != 0:
    #   raise errors.DiffEqError(f'We detect there are variable "{vars_in_eqs}" in the provided '
    #                            f'equations, while the user provided variables "{variables}" '
    #                            f'is not the same.')

    # finally
    self.eqs = eqs
    # self.variables = variables
    self.arg_keys = vars_in_eqs + ['t'] + all_arg_pars
    self.kwarg_keys = list(all_kwarg_pars.keys())
    self.kwargs = all_kwarg_pars
    parameters = [inspect.Parameter(vp, inspect.Parameter.POSITIONAL_OR_KEYWORD)
                  for vp in self.arg_keys]
    parameters.extend([inspect.Parameter(k,
                                         kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                                         default=all_kwarg_pars[k])
                       for k in self.kwarg_keys])
    signature = inspect.signature(eqs[0])
    self.__signature__ = signature.replace(parameters=parameters)
    self.__name__ = 'joint_eq'

  def __call__(self, *args, **kwargs):
    # format arguments
    params_in = Collector()
    for i, arg in enumerate(args):
      if i < len(self.arg_keys):
        params_in[self.arg_keys[i]] = arg
      else:
        params_in[self.kwarg_keys[i - len(self.arg_keys)]] = arg
    params_in.update(kwargs)

    # call equations
    results = []
    for i, eq in enumerate(self.eqs):
      r = eq(**{arg: params_in[arg] for arg in self.args_in_eqs[i]})
      if isinstance(r, (list, tuple)):
        results.extend(list(r))
      else:
        results.append(r)
    return results



# -*- coding: utf-8 -*-

import warnings

from brainpy import losses

__all__ = [
  'cross_entropy_loss',
  'l1_loos',
  'l2_loss',
  'l2_norm',
  'huber_loss',
  'mean_absolute_error',
  'mean_squared_error',
  'mean_squared_log_error',
]


def cross_entropy_loss(*args, **kwargs):
  """Cross entropy loss.

  .. deprecated:: 2.1.0
     Please use "brainpy.losses.cross_entropy_loss" instead.
  """
  warnings.warn('Please use "brainpy.losses.XXX" instead. '
                '"brainpy.math.losses.XXX" is deprecated since version 2.0.3. ',
                DeprecationWarning)
  return losses.cross_entropy_loss(*args, **kwargs)


def l1_loos(*args, **kwargs):
  """L1 loss.

  .. deprecated:: 2.1.0
     Please use "brainpy.losses.l1_loss" instead.
  """
  warnings.warn('Please use "brainpy.losses.XXX" instead. '
                '"brainpy.math.losses.XXX" is deprecated since version 2.0.3. ',
                DeprecationWarning)
  return losses.l1_loos(*args, **kwargs)


def l2_loss(*args, **kwargs):
  """L2 loss.

  .. deprecated:: 2.1.0
     Please use "brainpy.losses.l2_loss" instead.
  """
  warnings.warn('Please use "brainpy.losses.XXX" instead. '
                '"brainpy.math.losses.XXX" is deprecated since version 2.0.3. ',
                DeprecationWarning)
  return losses.l2_loss(*args, **kwargs)


def l2_norm(*args, **kwargs):
  """L2 normal.

  .. deprecated:: 2.1.0
     Please use "brainpy.losses.l2_norm" instead.
  """
  warnings.warn('Please use "brainpy.losses.XXX" instead. '
                '"brainpy.math.losses.XXX" is deprecated since version 2.0.3. ',
                DeprecationWarning)
  return losses.l2_norm(*args, **kwargs)


def huber_loss(*args, **kwargs):
  """Huber loss.

  .. deprecated:: 2.1.0
     Please use "brainpy.losses.huber_loss" instead.
  """
  warnings.warn('Please use "brainpy.losses.XXX" instead. '
                '"brainpy.math.losses.XXX" is deprecated since version 2.0.3. ',
                DeprecationWarning)
  return losses.huber_loss(*args, **kwargs)


def mean_absolute_error(*args, **kwargs):
  """mean absolute error loss.

  .. deprecated:: 2.1.0
     Please use "brainpy.losses.mean_absolute_error" instead.
  """
  warnings.warn('Please use "brainpy.losses.XXX" instead. '
                '"brainpy.math.losses.XXX" is deprecated since version 2.0.3. ',
                DeprecationWarning)
  return losses.mean_absolute_error(*args, **kwargs)


def mean_squared_error(*args, **kwargs):
  """Mean squared error loss.

  .. deprecated:: 2.1.0
     Please use "brainpy.losses.mean_squared_error" instead.
  """
  warnings.warn('Please use "brainpy.losses.XXX" instead. '
                '"brainpy.math.losses.XXX" is deprecated since version 2.0.3. ',
                DeprecationWarning)
  return losses.mean_squared_error(*args, **kwargs)


def mean_squared_log_error(*args, **kwargs):
  """Mean squared log error loss.

  .. deprecated:: 2.1.0
     Please use "brainpy.losses.mean_squared_log_error" instead.
  """
  warnings.warn('Please use "brainpy.losses.XXX" instead. '
                '"brainpy.math.losses.XXX" is deprecated since version 2.0.3. ',
                DeprecationWarning)
  return losses.mean_squared_log_error(*args, **kwargs)



# -*- coding: utf-8 -*-


from typing import Dict, Sequence

import brainpy.math as bm
from brainpy.base.base import Base
from brainpy.errors import DiffEqError
from brainpy.integrators.constants import DT
from brainpy.tools.checking import check_float, check_dict_data

__all__ = [
  'Integrator',
]


class AbstractIntegrator(Base):
  """Basic Integrator Class."""

  # func_name
  # derivative
  # code_scope
  #

  def __call__(self, *args, **kwargs):
    raise NotImplementedError


class Integrator(AbstractIntegrator):
  """Basic Integrator Class."""

  def __init__(
      self,
      variables: Sequence[str],
      parameters: Sequence[str],
      arguments: Sequence[str],
      dt: float,
      name: str = None,
      state_delays: Dict[str, bm.AbstractDelay] = None,
  ):
    super(Integrator, self).__init__(name=name)

    self._dt = dt
    check_float(dt, 'dt', allow_none=False, allow_int=True)
    self._variables = list(variables)  # variables
    self._parameters = list(parameters)  # parameters
    self._arguments = list(arguments) + [f'{DT}={self._dt}', ]  # arguments
    self._integral = None  # integral function

    # state delays
    self._state_delays = dict()
    if state_delays is not None:
      check_dict_data(state_delays, key_type=str, val_type=(bm.TimeDelay, bm.LengthDelay))
      for key, delay in state_delays.items():
        if key not in self.variables:
          raise DiffEqError(f'"{key}" is not defined in the variables: {self.variables}')
        self._state_delays[key] = delay
    self.register_implicit_nodes(self._state_delays)

  @property
  def dt(self):
    """The numerical integration precision."""
    return self._dt

  @dt.setter
  def dt(self, value):
    raise ValueError('Cannot set "dt" by users.')

  @property
  def variables(self):
    """The variables defined in the differential equation."""
    return self._variables

  @variables.setter
  def variables(self, values):
    raise ValueError('Cannot set "variables" by users.')

  @property
  def arg_names(self):
    return self._variables + self._parameters + [DT]

  @property
  def parameters(self):
    """The parameters defined in the differential equation."""
    return self._parameters

  @parameters.setter
  def parameters(self, values):
    raise ValueError('Cannot set "parameters" by users.')

  @property
  def arguments(self):
    """All arguments when calling the numer integrator of the differential equation."""
    return self._arguments

  @arguments.setter
  def arguments(self, values):
    raise ValueError('Cannot set "arguments" by users.')

  @property
  def integral(self):
    """The integral function."""
    return self._integral

  @integral.setter
  def integral(self, f):
    self.set_integral(f)

  def set_integral(self, f):
    """Set the integral function."""
    if not callable(f):
      raise ValueError(f'integral function must be a callable function, '
                       f'but we got {type(f)}: {f}')
    self._integral = f

  @property
  def state_delays(self):
    """State delays."""
    return self._state_delays

  @state_delays.setter
  def state_delays(self, value):
    raise ValueError('Cannot set "state_delays" by users.')

  def __call__(self, *args, **kwargs):
    assert self.integral is not None, 'Please build the integrator first.'

    # check arguments
    for i, arg in enumerate(args):
      kwargs[self.arg_names[i]] = arg

    # integral
    new_vars = self.integral(**kwargs)
    if len(self.variables) == 1:
      dict_vars = {self.variables[0]: new_vars}
    else:
      dict_vars = {k: new_vars[i] for i, k in enumerate(self.variables)}

    # update state delay variables
    dt = kwargs.pop(DT, self.dt)
    for key, delay in self.state_delays.items():
      if isinstance(delay, bm.LengthDelay):
        delay.update(dict_vars[key])
      elif isinstance(delay, bm.TimeDelay):
        delay.update(kwargs['t'] + dt, dict_vars[key])
      else:
        raise ValueError('Unknown delay variable. We only supports '
                         'brainpy.math.LengthDelay, brainpy.math.TimeDelay. '
                         f'While we got {delay}')

    return new_vars



# -*- coding: utf-8 -*-

import _thread as thread
import threading
from typing import Optional, Tuple

import numpy as np
from jax import lax
from jax.experimental import host_callback
from tqdm.auto import tqdm

__all__ = [
  'to_size',
  'size2num',
  'timeout',
  'init_progress_bar',
]


def size2num(size):
  if isinstance(size, int):
    return size
  elif isinstance(size, (tuple, list)):
    a = 1
    for b in size:
      a *= b
    return a
  else:
    raise ValueError(f'Do not support type {type(size)}: {size}')


def to_size(x) -> Optional[Tuple[int]]:
  if isinstance(x, (tuple, list)):
    return tuple(x)
  if isinstance(x, int):
    return (x, )
  if x is None:
    return x
  raise ValueError(f'Cannot make a size for {x}')


def timeout(s):
  """Add a timeout parameter to a function and return it.

  Parameters
  ----------
  s : float
      Time limit in seconds.

  Returns
  -------
  func : callable
      Functional results. Or, raise an error of KeyboardInterrupt.
  """

  def outer(fn):
    def inner(*args, **kwargs):
      timer = threading.Timer(s, thread.interrupt_main)
      timer.start()
      try:
        result = fn(*args, **kwargs)
      finally:
        timer.cancel()
      return result

    return inner

  return outer


def init_progress_bar(duration, dt, report=0.01, message=None):
  """Setup a progress bar."""
  if message is None:
    message = f"Running a duration of {duration}"

  num_samples = int(duration / dt)
  print_rate = int(duration * report / dt)
  remainder = num_samples % print_rate

  tqdm_bars = {}

  def _define_tqdm(arg, transform):
    tqdm_bars[0] = tqdm(np.arange(0, duration, dt))
    tqdm_bars[0].set_description(message, refresh=False)

  def _update_tqdm(num_processed, transform):
    tqdm_bars[0].update(num_processed * dt)

  def _update_progress_bar(num_iter):
    _ = lax.cond(
      num_iter == 0,
      lambda _: host_callback.id_tap(_define_tqdm, None, result=num_iter),
      lambda _: num_iter,
      operand=None,
    )

    _ = lax.cond(
      # update tqdm every multiple of `print_rate` except at the end
      (num_iter % print_rate == 0) & (num_iter != num_samples - remainder),
      lambda _: host_callback.id_tap(_update_tqdm, print_rate, result=num_iter),
      lambda _: num_iter,
      operand=None,
    )

    _ = lax.cond(
      # update tqdm by `remainder`
      num_iter == num_samples - remainder,
      lambda _: host_callback.id_tap(_update_tqdm, remainder, result=num_iter),
      lambda _: num_iter,
      operand=None,
    )

  def _close_tqdm(arg, transform):
    tqdm_bars[0].close()

  def close_tqdm(iter_num):
    return lax.cond(
      iter_num == num_samples - 1,
      lambda _: host_callback.id_tap(_close_tqdm, None, result=None),
      lambda _: None,
      operand=None,
    )

  def _progress_bar(iter_num):
    _update_progress_bar(iter_num)
    close_tqdm(iter_num)

  return _progress_bar



# -*- coding: utf-8 -*-

__version__ = "2.1.7"


try:
  import jaxlib
  del jaxlib
except ModuleNotFoundError:
  raise ModuleNotFoundError(
    'Please install jaxlib. See '
    'https://brainpy.readthedocs.io/en/latest/quickstart/installation.html#dependency-2-jax '
    'for installation instructions.'
  )


# fundamental modules
from . import errors, tools, check


# "base" module
from . import base
from .base.base import Base
from .base.collector import Collector, TensorCollector


# math foundation
from . import math


# toolboxes
from . import connect, initialize, optimizers, measure, losses, datasets, inputs


# numerical integrators
from . import integrators
from .integrators import ode
from .integrators import sde
from .integrators import dde
from .integrators import fde
from .integrators.ode import odeint
from .integrators.sde import sdeint
from .integrators.dde import ddeint
from .integrators.fde import fdeint
from .integrators.joint_eq import JointEq


# dynamics simulation
from . import dyn


# neural networks modeling
from . import nn


# running
from . import running


# automatic dynamics analysis
from . import analysis


# "visualization" module, will be removed soon
from .visualization import visualize


# compatible interface
from .compat import *  # compat


# convenient access
conn = connect
init = initialize
optim = optimizers
